"""Test features of lightkurve that interact with the data archive at MAST.

Note: if you have the `pytest-remotedata` package installed, then tests flagged
with the `@pytest.mark.remote_data` decorator below will only run if the
`--remote-data` argument is passed to py.test.  This allows tests to pass
if no internet connection is available.
"""
import os
import pytest

from numpy.testing import assert_almost_equal, assert_array_equal
import tempfile
from requests import HTTPError

from astropy.coordinates import SkyCoord
import astropy.units as u
from astropy.table import Table

from ..utils import LightkurveWarning, LightkurveError
from ..search import search_lightcurve, search_targetpixelfile, \
                     search_tesscut, SearchResult, SearchError, log
from .. import KeplerTargetPixelFile, TessTargetPixelFile, TargetPixelFileCollection


@pytest.mark.remote_data
def test_search_targetpixelfile():
    # EPIC 210634047 was observed twice in long cadence
    assert(len(search_targetpixelfile('EPIC 210634047', mission='K2').table) == 2)
    # ...including Campaign 4
    assert(len(search_targetpixelfile('EPIC 210634047', mission='K2', campaign=4).table) == 1)
    # KIC 11904151 (Kepler-10) was observed in LC in 15 Quarters
    assert(len(search_targetpixelfile('KIC 11904151', mission='Kepler', cadence='long').table) == 15)
    # ...including quarter 11 but not 12:
    assert(len(search_targetpixelfile('KIC 11904151', mission='Kepler', cadence='long', quarter=11).unique_targets) == 1)
    assert(len(search_targetpixelfile('KIC 11904151', mission='Kepler', cadence='long', quarter=12).table) == 0)
    search_targetpixelfile('KIC 11904151', quarter=11, cadence='long').download()
    # with mission='TESS', it should return TESS observations
    tic = 'TIC 273985862'  # Has been observed in multiple sectors including 1
    assert(len(search_targetpixelfile(tic, mission='TESS').table) > 1)
    assert(len(search_targetpixelfile(tic, mission='TESS', sector=1, radius=100).table) == 2)
    search_targetpixelfile(tic, mission='TESS', sector=1).download()
    assert(len(search_targetpixelfile("pi Mensae", sector=1).table) == 1)
    # Issue #445: indexing with -1 should return the last index of the search result
    assert(len(search_targetpixelfile("pi Men")[-1]) == 1)


@pytest.mark.remote_data
def test_search_split_campaigns():
    """Searches should should work for split campaigns.

    K2 Campaigns 9, 10, and 11 were split into two halves for various technical
    reasons (C91=C9a, C92=C9b, C101=C10a, C102=C10b, C111=C11a, C112=C11b).
    We expect most targets from those campaigns to return two TPFs.
    """
    campaigns = [9, 10, 11]
    ids = ['EPIC 228162462', 'EPIC 228726301', 'EPIC 202975993']
    for c, idx in zip(campaigns, ids):
        search = search_targetpixelfile(idx, campaign=c, cadence='long').table
        assert(len(search) == 2)


@pytest.mark.remote_data
def test_search_lightcurve(caplog):
    # We should also be able to resolve it by its name instead of KIC ID
    assert(len(search_lightcurve('Kepler-10', mission='Kepler', cadence='long').table) == 15)
    # An invalid KIC/EPIC ID or target name should be dealt with gracefully
    search_lightcurve(-999)
    assert "Could not resolve" in caplog.text
    search_lightcurve("DOES_NOT_EXIST (UNIT TEST)")
    assert "Could not resolve" in caplog.text
    # If we ask for all cadence types, there should be four Kepler files given
    assert(len(search_lightcurve('KIC 4914423', quarter=6, cadence='any').table) == 4)
    # ...and only one should have long cadence
    assert(len(search_lightcurve('KIC 4914423', quarter=6, cadence='long').table) == 1)
    # Should be able to resolve an ra/dec
    assert(len(search_lightcurve('297.5835, 40.98339', quarter=6).table) == 1)
    # Should be able to resolve a SkyCoord
    c = SkyCoord('297.5835 40.98339', unit=(u.deg, u.deg))
    search = search_lightcurve(c, quarter=6)
    assert(len(search.table) == 1)
    assert(len(search) == 1)
    # We should be able to download a light curve
    search.download()
    # The second call to download should use the local cache
    caplog.clear()
    caplog.set_level("DEBUG")
    search.download()
    assert "found in local cache" in caplog.text
    # with mission='TESS', it should return TESS observations
    tic = 'TIC 273985862'
    assert(len(search_lightcurve(tic, mission='TESS').table) > 1)
    assert(len(search_lightcurve(tic, mission='TESS', sector=1, radius=100).table) == 2)
    search_lightcurve(tic, mission='TESS', sector=1).download()
    assert(len(search_lightcurve("pi Mensae", sector=1).table) == 1)


@pytest.mark.remote_data
def test_search_tesscut():
    # Cutout by target name
    assert(len(search_tesscut("pi Mensae", sector=1).table) == 1)
    assert(len(search_tesscut("pi Mensae").table) > 1)
    # Cutout by TIC ID
    assert(len(search_tesscut('TIC 206669860', sector=2).table) == 1)
    # Cutout by RA, dec string
    search_string = search_tesscut('30.578761, -83.210593')
    # Cutout by SkyCoord
    c = SkyCoord('30.578761 -83.210593', unit=(u.deg, u.deg))
    search_coords = search_tesscut(c)
    # These should be identical
    assert(len(search_string.table) == len(search_coords.table))
    # The coordinates below are beyond the edge of the sector 4 (camera 1-4) FFI
    search_edge = search_tesscut('30.578761, 6.210593', sector=4)
    assert(len(search_edge.table) == 0)


@pytest.mark.remote_data
def test_search_tesscut_download(caplog):
    """Can we download TESS cutouts via `search_cutout().download()?"""
    try:
        ra, dec = 30.578761, -83.210593
        search_string = search_tesscut('{}, {}'.format(ra, dec), sector=[1, 12])
        # Make sure they can be downloaded with default size
        tpf = search_string[1].download()
        # Ensure the correct object has been returned
        assert(isinstance(tpf, TessTargetPixelFile))
        # Ensure default size is 5x5
        assert(tpf.flux[0].shape == (5, 5))
        assert(len(tpf.targetid) > 0)  # Regression test #473
        assert(tpf.sector == 12)  # Regression test #696
        # Ensure the WCS is valid (#434 regression test)
        center_ra, center_dec = tpf.wcs.all_pix2world([[2.5, 2.5]], 1)[0]
        assert_almost_equal(ra, center_ra, decimal=1)
        assert_almost_equal(dec, center_dec, decimal=1)
        # Download with different dimensions
        tpfc = search_string.download_all(cutout_size=4, quality_bitmask='hard')
        assert(isinstance(tpfc, TargetPixelFileCollection))
        assert(tpfc[0].quality_bitmask == 'hard')  # Regression test for #494
        assert(tpfc[0].sector == 1)  # Regression test #696
        assert(tpfc[1].sector == 12) # Regression test #696
        # Ensure correct dimensions
        assert(tpfc[0].flux[0].shape == (4, 4))
        # Download with rectangular dimennsions?
        rect_tpf = search_string[0].download(cutout_size=(3, 5))
        assert(rect_tpf.flux[0].shape == (3, 5))
        # If we ask for the exact same cutout, do we get it from cache?
        caplog.clear()
        log.setLevel("DEBUG")
        tpf_cached = search_string[0].download(cutout_size=(3, 5))
        assert "Cached file found." in caplog.text
    except HTTPError as exc:
        # TESSCut will occasionally return a "504 Gateway Timeout error" when
        # it is overloaded.  We don't want this to trigger a test failure.
        if "504" not in str(exc):
            raise exc


@pytest.mark.remote_data
def test_search_with_skycoord():
    """Can we pass both names, SkyCoord objects, and coordinate strings?"""
    sr_name = search_targetpixelfile("Kepler-10", mission='Kepler', cadence='long')
    assert len(sr_name) == 15  # Kepler-10 as observed during 15 quarters in long cadence
    # Can we search using a SkyCoord objects?
    sr_skycoord = search_targetpixelfile(SkyCoord.from_name("Kepler_10"), mission='Kepler', cadence='long')
    assert_array_equal(sr_name.table['productFilename'], sr_skycoord.table['productFilename'])
    # Can we search using a string of "ra dec" decimals?
    sr_decimal = search_targetpixelfile("285.67942179 +50.24130576", mission='Kepler', cadence='long')
    assert_array_equal(sr_name.table['productFilename'], sr_decimal.table['productFilename'])
    # Can we search using a sexagesimal string?
    sr_sexagesimal = search_targetpixelfile("19:02:43.1 +50:14:28.7", mission='Kepler', cadence='long')
    assert_array_equal(sr_name.table['productFilename'], sr_sexagesimal.table['productFilename'])
    # Can we search using the KIC ID?
    sr_kic = search_targetpixelfile('KIC 11904151', mission='Kepler', cadence='long')
    assert_array_equal(sr_name.table['productFilename'], sr_kic.table['productFilename'])


@pytest.mark.remote_data
def test_searchresult():
    sr = search_lightcurve('Kepler-10', mission='Kepler')
    assert len(sr) == len(sr.table)  # Tests SearchResult.__len__
    assert len(sr[2:7]) == 5  # Tests SearchResult.__get__
    assert len(sr[2]) == 1
    assert "kplr" in sr.__repr__()
    assert "kplr" in sr._repr_html_()


@pytest.mark.remote_data
def test_month():
    # In short cadence, if we specify both quarter and month
    sr = search_targetpixelfile('Kepler-10', quarter=11, month=1, cadence='short')
    assert(len(sr) == 1)
    sr = search_targetpixelfile('Kepler-10', quarter=11, month=[1, 3], cadence='short')
    assert(len(sr) == 2)


@pytest.mark.remote_data
def test_collections():
    # TargetPixelFileCollection class
    assert(len(search_targetpixelfile('EPIC 205998445', mission='K2',radius=900).table) == 4)
    # LightCurveFileCollection class with set targetlimit
    assert(len(search_lightcurve('EPIC 205998445', mission='K2', radius=900, limit=3).download_all()) == 3)
    # if fewer targets are found than targetlimit, should still download all available
    assert(len(search_targetpixelfile('EPIC 205998445', mission='K2', radius=900, limit=6).table) == 4)
    # if download() is used when multiple files are available, should only download 1
    with pytest.warns(LightkurveWarning, match='4 files available to download'):
        assert(isinstance(search_targetpixelfile('EPIC 205998445', mission='K2', radius=900).download(),
                          KeplerTargetPixelFile))


@pytest.mark.remote_data
def test_properties():
    c = SkyCoord('297.5835 40.98339', unit=(u.deg, u.deg))
    assert_almost_equal(search_targetpixelfile(c, quarter=6).ra, 297.5835)
    assert_almost_equal(search_targetpixelfile(c, quarter=6).dec, 40.98339)
    assert(len(search_targetpixelfile(c, quarter=6).target_name) == 1)
    assert(len(search_targetpixelfile(c, quarter=6).obsid) == 1)


@pytest.mark.remote_data
def test_source_confusion():
    # Regression test for issue #148.
    # When obtaining the TPF for target 6507433, @benmontet noticed that
    # a target 4 arcsec away was returned instead.
    # See https://github.com/KeplerGO/lightkurve/issues/148
    desired_target = "KIC 6507433"
    tpf = search_targetpixelfile(desired_target, quarter=8).download()
    assert tpf.targetid == 6507433


def test_empty_searchresult():
    """Does an empty SearchResult behave gracefully?"""
    sr = SearchResult(Table())
    assert len(sr) == 0
    str(sr)
    with pytest.warns(LightkurveWarning, match='empty search'):
        sr.download()
    with pytest.warns(LightkurveWarning, match='empty search'):
        sr.download_all()


@pytest.mark.remote_data
def test_issue_472():
    """Regression test for https://github.com/KeplerGO/lightkurve/issues/472"""
    # The line below previously threw an exception because the target was not
    # observed in Sector 2; we're always expecting a SearchResult object (empty
    # or not) rather than an exception.
    # Whether or not this SearchResult is empty has changed over the years,
    # because the target is only ~15 pixels beyond the FFI edge and the accuracy
    # of the FFI footprint polygons at the MAST portal have changed at times.
    search = search_tesscut("TIC41336498", sector=2)
    assert isinstance(search, SearchResult)


@pytest.mark.remote_data
def test_corrupt_download_handling():
    """When a corrupt file exists in the cache, make sure the user receives
    a helpful error message.

    This is a regression test for #511.
    """
    with tempfile.TemporaryDirectory() as tmpdirname:
        # Pretend a corrupt file exists at the expected cache location
        expected_dir = os.path.join(tmpdirname,
                                   "mastDownload",
                                   "Kepler",
                                   "kplr011904151_lc_Q111111110111011101")
        expected_fn = os.path.join(expected_dir, "kplr011904151-2010009091648_lpd-targ.fits.gz")
        os.makedirs(expected_dir)
        open(expected_fn, 'w').close()  # create "corrupt" i.e. empty file
        with pytest.raises(LightkurveError) as err:
            search_targetpixelfile("Kepler-10", quarter=4, cadence='long').download(download_dir=tmpdirname)
        assert "may be corrupt" in err.value.args[0]


@pytest.mark.remote_data
def test_indexerror_631():
    """Regression test for #631; avoid IndexError."""
    # This previously triggered an exception:
    result = search_lightcurve("KIC 8462852", sector=15, radius=1)
    assert len(result) == 1


@pytest.mark.remote_data
def test_name_resolving_regression_764():
    """Due to a bug, MAST resolved "EPIC250105131" to a different position than
    "EPIC 250105131". This regression test helps us verify that the bug does
    not re-appear. Details: https://github.com/KeplerGO/lightkurve/issues/764
    """
    from astroquery.mast import MastClass
    c1 = MastClass().resolve_object(objectname="EPIC250105131")
    c2 = MastClass().resolve_object(objectname="EPIC 250105131")
    assert c1.separation(c2).to("arcsec").value < 0.1


@pytest.mark.remote_data
def test_overlapping_targets_718():
    """Regression test for #718."""
    # Searching for the following targets without radius should only return
    # the requested targets, not their overlapping neighbors.
    targets = ['KIC 5112705', 'KIC 10058374', 'KIC 5385723']
    for target in targets:
        search = search_lightcurve(target, quarter=11)
        assert len(search) == 1
        assert search.target_name[0] == f'kplr{target[4:].zfill(9)}'

    # When using `radius=1` we should also retrieve the overlapping targets
    search = search_lightcurve('KIC 5112705', quarter=11, radius=1*u.arcsec)
    assert len(search) > 1

    # Searching by `target_name` should not preven a KIC identifier to work
    # in a TESS data search
    search = search_targetpixelfile('KIC 8462852', mission='TESS', sector=15)
    assert len(search) == 1


@pytest.mark.remote_data
def test_tesscut_795():
    """Regression test for #795: make sure the __repr__.of a TESSCut
    SearchResult works."""
    str(search_tesscut('KIC 8462852'))  # This raised a KeyError


@pytest.mark.remote_data
def test_download_flux_column():
    """Can we pass reader keyword arguments to the download method?"""
    lc = search_lightcurve("Pi Men", sector=12).download(flux_column='sap_flux')
    assert_array_equal(lc.flux, lc.sap_flux)


@pytest.mark.remote_data
def test_cadence_filtering():
    """Can we pass "fast", "short", exposure time to the cadence argument?"""
    # Try `cadence="fast"`
    res = search_lightcurve("AU Mic", sector=27, cadence="fast")
    assert(len(res) == 1)
    assert res.table['t_exptime'][0] == 20
    # Try `cadence="short"`
    res = search_lightcurve("AU Mic", sector=27, cadence="short")
    assert(len(res) == 1)
    assert res.table['t_exptime'][0] == 120
    # Try `cadence=20`
    res = search_lightcurve("AU Mic", sector=27, cadence=20)
    assert(len(res) == 1)
    assert res.table['t_exptime'][0] == 20
    assert "fast" in res.table['productFilename'][0]


@pytest.mark.remote_data
def test_ffi_hlsp():
    """Can SPOC, QLP (FFI), and TESS-SPOC (FFI) light curves be accessed?"""
    search = search_lightcurve("TrES-2 b", mission="tess", author="any", sector=26)  # aka TOI 2140.01
    assert "QLP" in search.table["author"]
    assert "TESS-SPOC" in search.table["author"]
    assert "SPOC" in search.table["author"]
    # tess-spoc also products tpfs
    search = search_targetpixelfile("TrES-2 b", mission="tess", author="any", sector=26)
    assert "TESS-SPOC" in search.table["author"]
    assert "SPOC" in search.table["author"]


@pytest.mark.remote_data
def test_qlp_ffi_lightcurve():
    """Can we search and download an MIT QLP FFI light curve?"""
    search = search_lightcurve("TrES-2 b", sector=26, author="qlp")
    assert len(search) == 1
    assert search.author[0] == "QLP"
    assert search.t_exptime[0] == 30*u.minute  # Sector 26 had 30-minute FFIs
    lc = search.download()
    all(lc.flux == lc.kspsap_flux)


@pytest.mark.remote_data
def test_spoc_ffi_lightcurve():
    """Can we search and download a SPOC FFI light curve?"""
    search = search_lightcurve("TrES-2 b", sector=26, author="tess-spoc")
    assert len(search) == 1
    assert search.author[0] == "TESS-SPOC"
    assert search.t_exptime[0] == 30*u.minute  # Sector 26 had 30-minute FFIs
    lc = search.download()
    all(lc.flux == lc.pdcsap_flux)



"""Implements the abstract `Corrector` base class.
"""
from abc import ABC, abstractmethod

import matplotlib
import numpy as np

from .. import LightCurve
from .metrics import overfit_metric_lombscargle, underfit_metric_neighbors


class Corrector(ABC):
    """Abstract base class documenting the required structure of classes
    designed to remove systematic noise from light curves.

    Attributes
    ----------
    original_lc : LightCurve
        The uncorrected light curve.  Must be passed into (or computed by) the
        constructor method.
    corrected_lc : LightCurve
        Corrected light curve. Must be updated upon each call to the `correct()` method.
    cadence_mask : np.array of dtype=bool
        Boolean array with the same length as `original_lc`.
        True indicates that a cadence should be used to fit the noise model.
        By setting certain cadences to False, users can exclude those cadences
        from informing the noise model, which will help prevent the overfitting
        of those signals (e.g. exoplanet transits).
        By default, the cadence mask is True across all cadences.

    Methods
    -------
    __init__()
        Accepts all the data required to execute the correction.
        The constructor must set the `original_lc` attribute.
    correct() -> LightCurve
        Executes the correction, optionally accepting meaningful parameters that
        can be used to modify the way the correction is applied.
        This method must set or update the `corrected_lc` attribute on each run.
    diagnose() -> matplotlib.axes.Axes
        Creates plots to elucidate the user's most recent call to `correct()`.
    """

    @property
    def original_lc(self) -> LightCurve:
        if hasattr(self, "_original_lc"):
            return self._original_lc
        else:
            raise AttributeError("`original_lc` has not been instantiated yet.")

    @original_lc.setter
    def original_lc(self, original_lc):
        self._original_lc = original_lc

    @property
    def corrected_lc(self) -> LightCurve:
        if hasattr(self, "_corrected_lc"):
            return self._corrected_lc
        else:
            raise AttributeError(
                "You need to call the `correct()` method "
                "before you can access `corrected_lc`."
            )

    @corrected_lc.setter
    def corrected_lc(self, corrected_lc):
        self._corrected_lc = corrected_lc

    @property
    def cadence_mask(self) -> np.array:
        if not hasattr(self, "_cadence_mask"):
            self._cadence_mask = np.ones(len(self.original_lc), dtype=bool)
        return self._cadence_mask

    @cadence_mask.setter
    def cadence_mask(self, cadence_mask):
        self._cadence_mask = cadence_mask

    def __init__(self, original_lc: LightCurve) -> None:
        """Constructor method.

        The constructor shall:
        * accept all data required to run the correction (e.g. light curves,
        target pixel files, engineering data).
        * instantiate the `original_lc` property.
        """
        self.original_lc = original_lc

    @abstractmethod
    def correct(
        self, cadence_mask: np.array = None, optimize: bool = False
    ) -> LightCurve:
        """Returns a `LightCurve` from which systematic noise has been removed.

        This method shall:
        * accept meaningful parameters that can be used to tune the correction,
          including:
          - `optimize`: should an optimizer be used to tune the parameters?
          - `cadence_mask`: flags cadences to be used to fit the noise model.
        * store all parameters as object attributes (e.g. `self.optimize`, `self.cadence_mask`);
        * store helpful diagnostic information as object attributes;
        * store the result in the `self.corrected_lc` attribute;
        * return `self.corrected_lc`.
        """
        if cadence_mask:
            self.cadence_mask = cadence_mask
        # ... perform correction ...
        # self.corrected_lc = corrected_lc
        # return corrected_lc

    @abstractmethod
    def diagnose(self) -> matplotlib.axes.Axes:
        """Returns plots which elucidate the most recent call to `correct()`.

        This method shall plot useful diagnostic information which have been
        stored as object attributes during the most recent call to `correct()`.
        """
        pass

    def compute_overfit_metric(self, **kwargs) -> float:
        """ Measures the degree of over-fitting in the correction.

        See the docstring of `lightkurve.correctors.metrics.overfit_metric_lombscargle`
        for details.

        Returns
        -------
        overfit_metric : float
            A float in the range [0,1] where 0 => Bad, 1 => Good
        """
        return overfit_metric_lombscargle(
            # Ignore masked cadences in the computation
            self.original_lc[self.cadence_mask],
            self.corrected_lc[self.cadence_mask],
            **kwargs
        )

    def compute_underfit_metric(self, **kwargs) -> float:
        """Measures the degree of under-fitting the correction.

        See the docstring of `lightkurve.correctors.metrics.underfit_metric_neighbors`
        for details.

        Returns
        -------
        underfit_metric : float
            A float in the range [0,1] where 0 => Bad, 1 => Good
        """
        return underfit_metric_neighbors(self.corrected_lc[self.cadence_mask], **kwargs)



"""Defines the `SFFCorrector` class.

`SFFCorrector` enables systematics to be removed from light curves using the
Self Flat-Fielding (SFF) method described in Vanderburg and Johnson (2014).
"""
import logging
import warnings

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

from astropy.modeling import models, fitting
from astropy.units import Quantity

from . import DesignMatrix, DesignMatrixCollection, SparseDesignMatrixCollection
from .regressioncorrector import RegressionCorrector
from .designmatrix import create_spline_matrix, create_sparse_spline_matrix

from .. import MPLSTYLE
from ..utils import LightkurveWarning

log = logging.getLogger(__name__)

__all__ = ['SFFCorrector']


class SFFCorrector(RegressionCorrector):
    """Special case of `.RegressionCorrector` where the `.DesignMatrix` includes
    the target's centroid positions.

    The design matrix also contains columns representing a spline in time
    design to capture the intrinsic, long-term variability of the target.

    Parameters
    ----------
    lc : `.LightCurve`
        The light curve that needs to be corrected.
    """
    def __init__(self, lc):
        if getattr(lc, 'mission', '') == 'TESS':
            warnings.warn("The SFF correction method is not suitable for use "
                          "with TESS data, because the spacecraft motion does "
                          "not proceed along a consistent arc.",
                          LightkurveWarning)

        self.raw_lc = lc
        if lc.flux.unit.to_string() == '':
            lc = lc.copy()
        else:
            lc = lc.copy().normalize()

        # Setting these values as None so we don't get a value error if the
        # user tries to access them before "correct()"
        self.window_points = None
        self.windows = None
        self.bins = None
        self.timescale = None
        self.breakindex = None
        self.centroid_col = None
        self.centroid_row = None
        super(SFFCorrector, self).__init__(lc=lc)

    def __repr__(self):
        return 'SFFCorrector (LC: {})'.format(self.lc.targetid)

    def correct(self, centroid_col=None, centroid_row=None, windows=20, bins=5,
                timescale=1.5, breakindex=None, degree=3, restore_trend=False,
                additional_design_matrix=None, polyorder=None, sparse=False, **kwargs):
        """Find the best fit correction for the light curve.

        Parameters
        ----------
        centroid_col : np.ndarray of floats (optional)
            Array of centroid column positions. If ``None``, will use the
            `centroid_col` attribute of the input light curve by default.
        centroid_row : np.ndarray of floats (optional)
            Array of centroid row positions. If ``None``, will use the
            `centroid_row` attribute of the input light curve by default.
        windows : int
            Number of windows to split the data into to perform the correction.
            Default 20.
        bins : int
            Number of "knots" to place on the arclength spline. More bins will
            increase the number of knots, making the spline smoother in arclength.
            Default 10.
        timescale: float
            Time scale of the b-spline fit to the light curve in time, in units
            of input light curve time.
        breakindex : None, int or list of ints (optional)
            Optionally the user can break the light curve into sections. Set
            break index to either an index at which to break, or list of indicies.
        degree : int
            The degree of polynomials in the splines in time and arclength. Higher
            values will create smoother splines. Default 3.
        cadence_mask : np.ndarray of bools (optional)
            Mask, where True indicates a cadence that should be used.
        sigma : int (default 5)
            Standard deviation at which to remove outliers from fitting
        niters : int (default 5)
            Number of iterations to fit and remove outliers
        restore_trend : bool (default False)
            Whether to restore the long term spline trend to the light curve
        propagate_errors : bool (default False)
            Whether to propagate the uncertainties from the regression. Default is False.
            Setting to True will increase run time, but will sample from multivariate normal
            distribution of weights.
        additional_design_matrix : `~lightkurve.lightcurve.Correctors.DesignMatrix` (optional)
            Additional design matrix to remove, e.g. containing background vectors.
        polyorder : int
            Deprecated as of Lightkurve v1.4.  Use ``degree`` instead.

        Returns
        -------
        corrected_lc : `~lightkurve.lightcurve.LightCurve`
            Corrected light curve, with noise removed.
        """
        DMC, spline = DesignMatrixCollection, create_spline_matrix
        if sparse:
            DMC, spline = SparseDesignMatrixCollection, create_sparse_spline_matrix

        if polyorder is not None:
            warnings.warn("`polyorder` is deprecated and no longer used, "
                          "please use the `degree` keyword instead.",
                          LightkurveWarning)

        if centroid_col is None:
            self.lc = self.lc.remove_nans(column='centroid_col')
            centroid_col = self.lc.centroid_col
        if centroid_row is None:
            self.lc = self.lc.remove_nans(column='centroid_row')
            centroid_row = self.lc.centroid_row

        if np.any([~np.isfinite(centroid_row), ~np.isfinite(centroid_col)]):
            raise ValueError('Centroids contain NaN values.')

        self.window_points = _get_window_points(centroid_col, centroid_row,
                                                windows, breakindex=breakindex)
        self.windows = windows
        self.bins = bins
        self.timescale = timescale
        self.breakindex = breakindex
        self.arclength = _estimate_arclength(centroid_col, centroid_row)

        lower_idx = np.asarray(np.append(0, self.window_points), int)
        upper_idx = np.asarray(np.append(self.window_points, len(self.lc.time)), int)

        dms = []
        for idx, a, b in zip(range(len(lower_idx)), lower_idx, upper_idx):
            if isinstance(self.arclength, Quantity):
                ar = np.copy(self.arclength.value)
            else:
                ar = np.copy(self.arclength)
            knots = list(np.percentile(ar[a:b], np.linspace(0, 100, bins+1)[1:-1]))
            ar[~np.in1d(ar, ar[a:b])] = 0

            dm = spline(ar, knots=knots, degree=degree).copy()
            dm.columns = ['window{}_bin{}'.format(idx+1, jdx+1)
                                        for jdx in range(dm.shape[1])]

            # I'm putting VERY weak priors on the SFF motion vectors
            # (1e-6 is being added to prevent sigma from being zero)
            ps = np.ones(dm.shape[1]) * 10000 * self.lc[a:b].flux.std() + 1e-6
            dm.prior_sigma = ps
            dms.append(dm)

        sff_dm = DMC(dms).to_designmatrix(name='sff')#.standardize()



        # long term
        n_knots = int((self.lc.time.value[-1] - self.lc.time.value[0])/timescale)

        s_dm = spline(self.lc.time.value, n_knots=n_knots, name='spline')
        means = [np.average(chunk) for chunk in np.array_split(self.lc.flux, n_knots)]
#        means = [np.average(self.lc.flux, weights=s_dm.values[:, idx]) for idx in range(s_dm.shape[1])]
        s_dm.prior_mu = np.asarray(means)

        # I'm putting WEAK priors on the spline that it must be around 1
        s_dm.prior_sigma = np.ones(len(s_dm.prior_mu)) * 1000 * self.lc.flux.std() + 1e-6

        # additional
        if additional_design_matrix is not None:
            if not isinstance(additional_design_matrix, DesignMatrix):
                raise ValueError('`additional_design_matrix` must be a DesignMatrix object.')
            self.additional_design_matrix = additional_design_matrix
            dm = DMC([s_dm,
                                         sff_dm,
                                         additional_design_matrix])
        else:
            dm = DMC([s_dm, sff_dm])

        # correct
        clc = super(SFFCorrector, self).correct(dm, **kwargs)

        # clean
        if restore_trend:
            trend = self.diagnostic_lightcurves['spline'].flux
            clc += trend - np.nanmedian(trend)
        clc *= self.raw_lc.flux.mean()

        return clc

    def diagnose(self):
        """Returns a diagnostic plot which visualizes what happened during the
        most recent call to `correct()`."""
        axs = self._diagnostic_plot()
        for t in self.window_points:
            axs[0].axvline(self.lc.time.value[t], color='r', ls='--', alpha=0.3)

    def diagnose_arclength(self):
        """Returns a diagnostic plot which visualizes arclength vs flux
        from most recent call to `correct()`."""

        max_plot = 5
        with plt.style.context(MPLSTYLE):
            _, axs = plt.subplots(int(np.ceil(self.windows/max_plot)), max_plot,
                                  figsize=(10, int(np.ceil(self.windows/max_plot)*2)),
                                  sharex=True, sharey=True)
            axs = np.atleast_2d(axs)
            axs[0, 2].set_title('Arclength Plot/Window')
            plt.subplots_adjust(hspace=0, wspace=0)

            lower_idx = np.asarray(np.append(0, self.window_points), int)
            upper_idx = np.asarray(np.append(self.window_points, len(self.lc.time)), int)
            if hasattr(self, 'additional_design_matrix'):
                name = self.additional_design_matrix.name
                f = (self.lc.flux - self.diagnostic_lightcurves['spline'].flux
                            - self.diagnostic_lightcurves[name].flux)
            else:
                f = (self.lc.flux - self.diagnostic_lightcurves['spline'].flux)

            m = self.diagnostic_lightcurves['sff'].flux

            idx, jdx = 0, 0
            for a, b in zip(lower_idx, upper_idx):
                ax = axs[idx, jdx]
                if jdx == 0:
                    ax.set_ylabel('Flux')

                ax.scatter(self.arclength[a:b], f[a:b], s=1, label='Data')
                ax.scatter(self.arclength[a:b][~self.cadence_mask[a:b]],
                           f[a:b][~self.cadence_mask[a:b]],
                           s=10, marker='x', c='r', label='Outliers')

                s = np.argsort(self.arclength[a:b])
                ax.scatter(self.arclength[a:b][s],
                           (m[a:b] - np.median(m[a:b]) + np.median(f[a:b]))[s],
                           c='C2', s=0.5, label='Model')
                jdx += 1
                if jdx >= max_plot:
                    jdx = 0
                    idx += 1
                if b == len(self.lc.time):
                    ax.legend()


######################
#  Helper functions  #
######################

def _get_centroid_dm(col, row, name='centroids'):
    """Returns a `.DesignMatrix` containing (col, row) centroid positions
    and transformations thereof.

    Parameters
    ----------
    col : np.ndarray
        centroid column
    row : np.ndarray
        centroid row
    name : str
        Name to pass to `.DesignMatrix` (default: 'centroids').

    Returns
    -------
    dm: np.ndarray
        Design matrix with shape len(c) x 10
    """
    data = [col, row,
            col**2, row**2,
            col**3, row**3,
            col*row,
            col**2 * row, col * row**2,
            col**2 * row**2]
    names = [r'col', r'row',
             r'col^2', r'row^2',
             r'col^3', r'row^3',
             r'col \times row',
             r'col^2 \times row', r'col \times row^2',
             r'col^2 \times row^2']
    df = pd.DataFrame(np.asarray(data).T, columns=names)
    return DesignMatrix(df, name=name)


def _get_thruster_firings(arclength):
    """Find locations where K2 fired thrusters

    Parameters
    ----------
    arc : np.ndarray
        arclength as a function of time

    Returns
    -------
    thrusters: np.ndarray of bools
        True at times where thrusters were fired.
    """
    if isinstance(arclength, Quantity):
        arc = np.copy(arclength.value)
    else:
        arc = np.copy(arclength)
    # Rate of change of rate of change of arclength wrt time
    d2adt2 = (np.gradient(np.gradient(arc)))
    # Fit a Gaussian, most points lie in a tight region, thruster firings are outliers
    g = models.Gaussian1D(amplitude=100, mean=0, stddev=0.01)
    fitter = fitting.LevMarLSQFitter()
    h = np.histogram(d2adt2[np.isfinite(d2adt2)], np.arange(-0.5, 0.5, 0.0001), density=True)
    xbins = h[1][1:] - np.median(np.diff(h[1]))
    g = fitter(g, xbins, h[0], weights=h[0]**0.5)

    # Depending on the orientation of the roll, it is hard to return
    # the point before the firing or the point after the firing.
    # This makes sure we always return the same value, no matter the roll orientation.
    def _start_and_end(start_or_end):
        """Find points at the start or end of a roll."""
        if start_or_end == 'start':
            thrusters = (d2adt2 < (g.stddev * -5)) & np.isfinite(d2adt2)
        if start_or_end == 'end':
            thrusters = (d2adt2 > (g.stddev * 5)) & np.isfinite(d2adt2)
        # Pick the best thruster in each cluster
        idx = np.array_split(np.arange(len(thrusters)),
                             np.where(np.gradient(np.asarray(thrusters, int)) == 0)[0])
        m = np.array_split(thrusters, np.where(np.gradient(np.asarray(thrusters, int)) == 0)[0])
        th = []
        for jdx, _ in enumerate(idx):
            if m[jdx].sum() == 0:
                th.append(m[jdx])
            else:
                th.append((np.abs(np.gradient(arc)[idx[jdx]]) == np.abs(np.gradient(arc)[idx[jdx]][m[jdx]]).max()) & m[jdx])
        thrusters = np.hstack(th)
        return thrusters

    # Get the start and end points
    thrusters = np.asarray([_start_and_end('start'), _start_and_end('end')])
    thrusters = thrusters.any(axis=0)

    # Take just the first point.
    thrusters = (np.gradient(np.asarray(thrusters, int)) >= 0) & thrusters
    return thrusters


def _get_window_points(centroid_col, centroid_row, windows, arclength=None, breakindex=None):
    """Returns indices where thrusters are fired.

    Parameters
    ----------
    lc : `.LightCurve` object
        Input light curve
    windows: int
        Number of windows to split the light curve into
    arc: np.ndarray
        Arclength for the roll motion
    breakindex: int
        Cadence where there is a natural break. Windows will be automatically put here.
    """

    if arclength is None:
        arclength = _estimate_arclength(centroid_col, centroid_row)

    # Validate break indices
    if isinstance(breakindex, int):
        breakindexes = [breakindex]
    if breakindex is None:
        breakindexes = []
    elif (breakindex[0] == 0) & (len(breakindex) == 1):
        breakindexes = []
    else:
        breakindexes = breakindex

    if not isinstance(breakindexes, list):
        raise ValueError('`breakindex` must be an int or a list')

    # If the user asks for break indices we should still return them,
    # even if there is only 1 window.
    if windows == 1:
        return breakindexes

    # Find evenly spaced window points
    dt = len(centroid_col) / windows
    lower_idx = np.append(0, breakindexes)
    upper_idx = np.append(breakindexes, len(centroid_col))
    window_points = np.hstack([np.asarray(np.arange(a, b, dt), int)
                              for a, b in zip(lower_idx, upper_idx)])

    # Get thruster firings
    thrusters = _get_thruster_firings(arclength)
    for b in breakindexes:
        thrusters[b] = True
    thrusters = np.where(thrusters)[0]

    # Find the nearest point to each thruster firing, unless it's a user supplied break point
    if len(thrusters) > 0:
        window_points = [thrusters[np.argmin(np.abs(thrusters - wp))] + 1
                             for wp in window_points
                             if wp not in breakindexes]
    window_points = np.unique(np.hstack([window_points, breakindexes]))

    # If the first or last windows are very short (<40% median window length),
    # then we add them to the second or penultimate window, respectively,
    # by removing their break points.
    median_length = np.median(np.diff(window_points))
    if window_points[0] < 0.4*median_length:
        window_points = window_points[1:]
    if window_points[-1] > (len(centroid_col) - 0.4*median_length):
        window_points = window_points[:-1]

    return np.asarray(window_points, dtype=int)


def _estimate_arclength(centroid_col, centroid_row):
    """Estimate the arclength given column and row centroid positions.

    We use the approximation that the arclength equals

        (row**2 + col**2)**0.5

    For this to work, row and column must be correlated not anticorrelated.
    """
    col = centroid_col - np.nanmin(centroid_col)
    row = centroid_row  - np.nanmin(centroid_row)
    # Force c to be correlated not anticorrelated
    if np.polyfit(col.data, row.data, 1)[0] < 0:
        col = np.nanmax(col) - col
    return (col**2 + row**2)**0.5



import os

from astropy.io import fits

from ... import PACKAGEDIR
from .. import detect_filetype


def test_detect_filetype():
    """Can we detect the correct filetype?"""
    k2_path = os.path.join(PACKAGEDIR, "tests", "data", "test-tpf-star.fits")
    tess_path = os.path.join(PACKAGEDIR, "tests", "data", "tess25155310-s01-first-cadences.fits.gz")
    assert detect_filetype(fits.open(k2_path)) == 'KeplerTargetPixelFile'
    assert detect_filetype(fits.open(tess_path)) == 'TessTargetPixelFile'



from __future__ import division, print_function

import numpy as np
from numpy.testing import assert_almost_equal

from ..lightcurve import LightCurve
from ..convenience import estimate_cdpp


def test_cdpp():
    """Tests the estimate_cdpp() convenience function which wraps
    `LightCurve.estimate_cdpp()`"""
    flux = np.random.normal(loc=1, scale=100e-6, size=10000)
    lc = LightCurve(time=np.arange(10000), flux=flux)
    assert_almost_equal(estimate_cdpp(flux), lc.estimate_cdpp())



"""Helper functions for estimating numax from periodograms."""
import numpy as np
from matplotlib import pyplot as plt

from astropy.convolution import convolve, Gaussian1DKernel
from astropy import units as u

from .. import MPLSTYLE
from . import utils
from .utils import SeismologyQuantity

__all__ = ['estimate_numax_acf2d', 'diagnose_numax_acf2d']


def estimate_numax_acf2d(periodogram, numaxs=None, window_width=None, spacing=None):
    """Estimates the peak of the envelope of seismic oscillation modes, numax,
    using an autocorrelation function.

    There are many papers on the topic of autocorrelation functions for
    estimating seismic parameters, including but not limited to:
    Roxburgh & Vorontsov (2006), Roxburgh (2009), Mosser & Appourchaux (2009),
    Huber et al. (2009), Verner & Roxburgh (2011) & Viani et al. (2019).

    We base this approach first and foremost off the 2D ACF numax estimation
    presented in Viani et al. (2019) and other papers above. A window of
    fixed width (either given by the user, 25 microhertz for Red Giants or
    250 microhertz for Main Sequence stars) is moved along the power
    spectrum, where the central frequency of the window moves in steps of 1
    microhertz (or given by the user as `spacing`) and evaluates the
    autocorrelation at each step.

    The correlation (numpy.correlate) is typically given as:

        C[x, y] = sum( x * conj(y) ) .

    The autocorrelation power of a full spectrum with itself is then

        C = sum(s * s),

    where s is a window of the signal-to-noise spectrum.
    Because of the method of this calculation, we need to first
    rescale the power by subtracting its mean, placing its mean around 0. This
    decreases the noise levels in the ACF, as the autocorrelation of the noise
    with itself will be close to zero.

    In order to evaluate where the correlation power is highest (indicative
    of the power excess of the modes) we calculate the Mean Collapsed
    Correlation (MCC, see Kiefer 2013, Viani et al. 2019) as

        MCC = (sum(|C|) - 1) / nlags ,

    where C is the autocorrelation power at a given central freqeuncy, and
    nlags is the number of lags in the autocorrelation.

    The MCC metric is covolved with an Astropy Gaussian 1D Kernel with a
    standard deviation of 1/5th of the window size to smooth it. The
    frequency that results in the highest value of the smoothed MCC is the
    detected numax.

    NOTE: This method is not robust against large peaks in the spectrum (due
    to e.g. spacecraft rotation), nor is it robust in the case of low signal
    to noise (such as for single sector TESS data). Exercise caution when
    using this module!

    NOTE: This function is intended for use with solar like Main Sequence
    and Red Giant Branch oscillators only.

    Parameters
    ----------
    numaxs : array-like
        An array of numaxs at which to evaluate the autocorrelation. If
        none is given, a sensible range will be chosen. If no units are
        given it is assumed to be in the same units as the periodogram
        frequency.
    window_width : int or float
        The width of the autocorrelation window around each central
        frequency in 'numaxs'. If none is given, a sensible value will be
        chosen. If no units are given it is assumed to be in the same units
        as the periodogram frequency.
    spacing : int or float
        The spacing between central frequencies (numaxs) at which the
        autocorrelation is evaluated. If none is given, a sensible value
        will be assumed. If no units are given it is assumed to be in the
        same units as the periodogram frequency.

    Returns
    -------
    numax : `.SeismologyQuantity`
        The numax of the periodogram. In the units of the periodogram object
        frequency.
    """
    # Detect whether the frequency grid is evenly-spaced
    if not periodogram._is_evenly_spaced():
        raise ValueError("the ACF 2D method requires that the periodogram "
                         "has a grid of uniformly spaced frequencies.")

    # Calculate the window_width size

    #C: What is this doing? Why have these values been picked? This function is slow.
    if window_width is None:
        if u.Quantity(periodogram.frequency[-1], u.microhertz) > u.Quantity(500., u.microhertz):
            window_width = u.Quantity(250., u.microhertz).to(periodogram.frequency.unit).value
        else:
            window_width = u.Quantity(25., u.microhertz).to(periodogram.frequency.unit).value

    # Calculate the spacing size
    if spacing is None:
        if u.Quantity(periodogram.frequency[-1], u.microhertz) > u.Quantity(500., u.microhertz):
            spacing = u.Quantity(10., u.microhertz).to(periodogram.frequency.unit).value
        else:
            spacing = u.Quantity(1., u.microhertz).to(periodogram.frequency.unit).value

    # Run some checks on the inputs
    window_width = u.Quantity(window_width, periodogram.frequency.unit).value
    spacing = u.Quantity(spacing, periodogram.frequency.unit).value
    if numaxs is None:
        numaxs = np.arange(np.ceil(np.nanmin(periodogram.frequency.value)) + window_width/2,
                    np.floor(np.nanmax(periodogram.frequency.value)) - window_width/2,
                    spacing)
    numaxs = u.Quantity(numaxs, periodogram.frequency.unit).value
    if not hasattr(numaxs, '__iter__'):
        numaxs = np.asarray([numaxs])

    fs = np.median(np.diff(periodogram.frequency.value))
    # Perform checks on spacing and window_width
    for var, label in zip([np.asarray(window_width), np.asarray(spacing)],
                          ['window_width', 'spacing']):
        if (var < fs).any():
            raise ValueError("You can't have {} smaller than the "
                            "frequency separation!".format(label))
        if (var > (periodogram.frequency[-1].value - periodogram.frequency[0].value)).any():
            raise ValueError("You can't have {} wider than the entire "
                            "power spectrum!".format(label))
        if (var < 0).any():
            raise ValueError("Please pass an entirely positive {}.".format(label))

    # Perform checks on numaxs
    if any(numaxs < fs):
        raise ValueError("A custom range of numaxs can not extend below "
                        "a single frequency bin.")
    if any(numaxs > np.nanmax(periodogram.frequency.value)):
        raise ValueError("A custom range of numaxs can not extend above "
                        "the highest frequency value in the periodogram.")

    # We want to find the numax which returns in the highest autocorrelation
    # power, rescaled based on filter width
    fs = np.median(np.diff(periodogram.frequency.value))

    metric = np.zeros(len(numaxs))
    acf2d = np.zeros([int(window_width/2/fs)*2,len(numaxs)])
    for idx, numax in enumerate(numaxs):
        acf = utils.autocorrelate(periodogram, numax, window_width=window_width, frequency_spacing=fs)      #Return the acf at this numax
        acf2d[:, idx] = acf                                     #Store the 2D acf
        metric[idx] = (np.sum(np.abs(acf)) - 1 ) / len(acf)  #Store the max acf power normalised by the length

    # Smooth the data to find the peak
    # Gaussian1D kernel takes a standard deviation in unitless indices. A stddev
    # of sqrt(len(numaxs) will result in a smoothing kernel that works for all
    # resolutions of numax.
    if len(numaxs) > 10:
        g = Gaussian1DKernel(stddev=np.sqrt(len(numaxs)))
        metric_smooth = convolve(metric, g, boundary='extend')
    else:
        metric_smooth = metric

    # The highest value of the metric corresponds to numax
    best_numax = numaxs[np.argmax(metric_smooth)]
    best_numax = u.Quantity(best_numax, periodogram.frequency.unit)

    # Create and return the object containing the result and diagnostics
    diagnostics = {'numaxs':numaxs, 'acf2d':acf2d, 'window_width':window_width,
                   'metric':metric, 'metric_smooth': metric_smooth}
    result = SeismologyQuantity(best_numax,
                                name="numax",
                                method="ACF2D",
                                diagnostics=diagnostics,
                                diagnostics_plot_method=diagnose_numax_acf2d)
    return result


def diagnose_numax_acf2d(numax, periodogram):
    """Returns a diagnostic plot which elucidates how numax was estimated.

    [1] The SNRPeriodogram plotted with a red line indicating the estimated
    numax value.

    [2] An image showing the 2D autocorrelation. On the y-axis is the
    frequency lag of the autocorrelation window. The width of the window is
    equal to `window_width`, and the spacing between lags is equal to
    `numax_spacing`. On the x-axis is the central frequency at which the
    autocorrelation was calculated. In the z-axis is the unitless
    autocorrelation power. Shown in red is the estimated numax.

    [3] The Mean Collapsed Correlation (MCC, see Viani et al. 2019) against
    central frequency at which the MCC was calculated. Shown in red is the
    estimated numax. Shown in blue is the MCC convolved with a Gaussian
    smoothing kernel with a standard deviation of 1/5th the window size.

    For details on the numax estimation, see the `estimate_numax()` function.
    The calculation performed is identical

    Parameters
    ----------
    numax : `.SeismologyResult` object
        The object returned by `estimate_numax_acf2d()`.

    Returns
    -------
    ax : `~matplotlib.axes.Axes`
        The matplotlib axes object.
    """
    with plt.style.context(MPLSTYLE):
        fig, ax = plt.subplots(3, sharex=True, figsize=(8.485, 12))
        periodogram.plot(ax=ax[0], label='')
        ax[0].axvline(numax.value, c='r', linewidth=2, alpha=.4,
                label='{} = {:7.5} {}'.format(r'$\nu_{\rm max}$', numax.value, periodogram.frequency.unit.to_string('latex')))
        ax[0].legend(loc='upper right')
        ax[0].set_xlabel('')
        ax[0].text(.05, .9, 'Input Power Spectrum',
                    horizontalalignment='left',
                    transform=ax[0].transAxes, fontsize=15)

        vmin = np.nanpercentile(numax.diagnostics['acf2d'], 5)
        vmax = np.nanpercentile(numax.diagnostics['acf2d'], 95)
        ax[1].pcolormesh(numax.diagnostics['numaxs'],
                         np.linspace(0, numax.diagnostics['window_width'],
                         num=numax.diagnostics['acf2d'].shape[0]),
                         numax.diagnostics['acf2d'],
                         cmap='Blues',
                         vmin=vmin,
                         vmax=vmax)
        ax[1].set_ylabel(r'Frequency lag [{}]'.format(periodogram.frequency.unit.to_string('latex')))
        ax[1].axvline(numax.value, c='r', linewidth=2, alpha=.4)
        ax[1].text(.05, .9, '2D AutoCorrelation',
                    horizontalalignment='left',
                    transform=ax[1].transAxes, fontsize=13)

        ax[2].plot(numax.diagnostics['numaxs'], numax.diagnostics['metric'])
        ax[2].plot(numax.diagnostics['numaxs'], numax.diagnostics['metric_smooth'],
                   lw=2, alpha=0.7, label='Smoothed Metric')
        ax[2].set_xlabel("Frequency [{}]".format(periodogram.frequency.unit.to_string('latex')))
        ax[2].set_ylabel(r'Correlation Metric')

        ax[2].axvline(numax.value, c='r', linewidth=2, alpha=.4)
        ax[2].text(.05, .9, 'Correlation Metric',
                            horizontalalignment='left',
                            transform=ax[2].transAxes, fontsize=13)
        ax[2].legend(loc='upper right')
        ax[2].set_xlim(numax.diagnostics['numaxs'][0], numax.diagnostics['numaxs'][-1])
        plt.subplots_adjust(hspace=0, wspace=0)
    return ax



import pytest
import warnings

import numpy as np
from numpy.testing import assert_array_equal
import pandas as pd

from .. import SparseDesignMatrix, SparseDesignMatrixCollection, DesignMatrix, DesignMatrixCollection
from ... import LightkurveWarning
from ..designmatrix import create_sparse_spline_matrix, create_spline_matrix


from scipy import sparse


def test_designmatrix_basics():
    """Can we create a design matrix from a dataframe?"""
    size, name = 10, 'testmatrix'
    df = pd.DataFrame({'vector1': np.ones(size),
                       'vector2': np.arange(size),
                       'vector3': np.arange(size)**2})
    X = sparse.csr_matrix(np.asarray(df))
    dm = SparseDesignMatrix(X, name=name, columns=['vector1', 'vector2', 'vector3'])
    assert dm.columns == ['vector1', 'vector2', 'vector3']
    assert dm.name == name
    assert dm.shape == (size, 3)
    dm.plot()
    dm.plot_priors()
    assert dm.append_constant().shape == (size, 4)  # expect one column more
    assert dm.pca(nterms=2).shape == (size, 2)      # expect one column less
    assert dm.split([5]).shape == (size, 6)        # expect double columns
    dm.__repr__()

    dm = SparseDesignMatrix(X, name=name, columns=['vector1', 'vector2', 'vector3'])
    dm.append_constant(inplace=True)
    assert dm.shape == (size, 4)  # expect one column more

    dm = SparseDesignMatrix(X, name=name, columns=['vector1', 'vector2', 'vector3'])
    dm.split([5], inplace=True)
    assert dm.shape == (size, 6)        # expect double columns


def test_split():
    """Can we split a design matrix correctly?"""
    X = sparse.csr_matrix(np.vstack([np.linspace(0, 9, 10), np.linspace(100, 109, 10)]).T)
    dm = SparseDesignMatrix(X, columns=['a', 'b'])
    # Do we retrieve the correct shape?
    assert dm.shape == (10, 2)
    assert dm.split(2).shape == (10, 4)
    assert dm.split([2,8]).shape == (10, 6)
    # Are the new areas padded with zeros?
    assert (dm.split([2,8]).values[2:, 0:2] == 0).all()
    assert (dm.split([2,8]).values[:8, 4:] == 0).all()
    # Are all the column names unique?
    assert len(set(dm.split(4).columns)) == 4


def test_standardize():
    """Verifies DesignMatrix.standardize()"""
    # A column with zero standard deviation remains unchanged
    X = sparse.csr_matrix(np.vstack([np.ones(10)]).T)
    dm = SparseDesignMatrix(X, columns=['const'])
    assert (dm.standardize()['const'] == dm['const']).all()
    # Normally-distributed columns will become Normal(0, 1)
    X = sparse.csr_matrix(np.vstack([ np.random.normal(loc=5, scale=3, size=100)]).T)
    dm = SparseDesignMatrix(X, columns=['normal'])

    assert np.round(np.mean(dm.standardize()['normal']), 3) == 0
    assert np.round(np.std(dm.standardize()['normal']), 1) == 1
    dm.standardize(inplace=True)

def test_pca():
    """Verifies DesignMatrix.pca()"""
    size = 10
    dm = DesignMatrix({'a':np.random.normal(10, 20, size),
                       'b':np.random.normal(40, 10, size),
                       'c':np.random.normal(60, 5, size)}).to_sparse()
    for nterms in [1, 2, 3]:
        assert dm.pca(nterms=nterms).shape == (size, nterms)


def test_collection_basics():
    """Can we create a design matrix collection?"""
    size = 5
    dm1 = DesignMatrix(np.ones((size, 1)), columns=['col1'], name='matrix1').to_sparse()
    dm2 = DesignMatrix(np.zeros((size, 2)), columns=['col2', 'col3'], name='matrix2').to_sparse()

    dmc = SparseDesignMatrixCollection([dm1, dm2])
    assert_array_equal(dmc['matrix1'].values, dm1.values)
    assert_array_equal(dmc['matrix2'].values, dm2.values)
    assert_array_equal(dmc.values, np.hstack((dm1.values, dm2.values)))
    dmc.plot()
    dmc.__repr__()

    dmc = dm1.collect(dm2)
    assert_array_equal(dmc['matrix1'].values, dm1.values)
    assert_array_equal(dmc['matrix2'].values, dm2.values)
    assert_array_equal(dmc.values, np.hstack((dm1.values, dm2.values)))

    """Can we create a design matrix collection when one is sparse?"""
    size = 5
    dm1 = DesignMatrix(np.ones((size, 1)), columns=['col1'], name='matrix1')
    dm2 = DesignMatrix(np.zeros((size, 2)), columns=['col2', 'col3'], name='matrix2').to_sparse()

    with warnings.catch_warnings():
        warnings.simplefilter("always")
        with pytest.warns(LightkurveWarning, match='Sparse matrices will be converted to dense matrices.'):
            dmc = DesignMatrixCollection([dm1, dm2])
            assert not np.any([sparse.issparse(d.X) for d in dmc])

    with warnings.catch_warnings():
        warnings.simplefilter("always")
        with pytest.warns(LightkurveWarning, match='Dense matrices will be converted to sparse matrices.'):
            dmc = SparseDesignMatrixCollection([dm1, dm2])
            assert np.all([sparse.issparse(d.X) for d in dmc])

    dmc.plot()
    dmc.__repr__()

    assert isinstance(dmc.to_designmatrix(), SparseDesignMatrix)


def test_designmatrix_rank():
    """Does DesignMatrix issue a low-rank warning when justified?"""
    warnings.simplefilter("always")

    # Good rank
    dm = DesignMatrix({'a': [1, 2, 3]}).to_sparse()
    assert dm.rank == 1
    dm.validate(rank=True)  # Should not raise a warning

    # Bad rank
    with pytest.warns(LightkurveWarning, match='rank'):
        dm = DesignMatrix({'a': [1, 2, 3], 'b': [1, 1, 1], 'c': [1, 1, 1],
                           'd': [1, 1, 1], 'e': [3, 4, 5]})
        dm.validate(rank=True) # Should raise a warning
    dm = dm.to_sparse()
    assert dm.rank == 2
    with pytest.warns(LightkurveWarning, match='rank'):
        dm.validate(rank=True)


def test_splines():
    """Do splines work as expected?"""
    # Dense and sparse splines should produce the same answer.
    x = np.linspace(0, 1, 100)
    spline_dense = create_spline_matrix(x, knots=[0.1, 0.3, 0.6, 0.9], degree=2)
    spline_sparse = create_sparse_spline_matrix(x, knots=[0.1, 0.3, 0.6, 0.9], degree=2)
    assert np.allclose(spline_dense.values, spline_sparse.values)
    assert isinstance(spline_dense, DesignMatrix)
    assert isinstance(spline_sparse, SparseDesignMatrix)



"""Defines `RegressionCorrector` to solve large linear regression problems
with user-defined Gaussian priors in a fast, analytical way.
"""
import logging
import warnings

from astropy.stats import sigma_clip
from astropy import units as u
import matplotlib.pyplot as plt
import numpy as np
from scipy.sparse import issparse, csr_matrix

from .corrector import Corrector
from .designmatrix import DesignMatrix, DesignMatrixCollection, \
                          SparseDesignMatrix, SparseDesignMatrixCollection
from ..lightcurve import LightCurve, MPLSTYLE


__all__ = ['RegressionCorrector']


log = logging.getLogger(__name__)


class RegressionCorrector(Corrector):
    """Remove noise using linear regression against a `.DesignMatrix`.

    .. math::

        \\newcommand{\\y}{\\mathbf{y}}
        \\newcommand{\\cov}{\\boldsymbol\Sigma_\y}
        \\newcommand{\\w}{\\mathbf{w}}
        \\newcommand{\\covw}{\\boldsymbol\Sigma_\w}
        \\newcommand{\\muw}{\\boldsymbol\mu_\w}
        \\newcommand{\\sigw}{\\boldsymbol\sigma_\w}
        \\newcommand{\\varw}{\\boldsymbol\sigma^2_\w}

    Given a column vector of data :math:`\y`
    and a design matrix of regressors :math:`X`,
    we will find the vector of coefficients :math:`\w`
    such that:

    .. math::

        \mathbf{y} = X\mathbf{w} + \mathrm{noise}

    We will assume that the model fits the data within Gaussian uncertainties:

    .. math::

        p(\y | \w) = \mathcal{N}(X\w, \cov)


    We make the regression robust by placing Gaussian priors on :math:`\w`:

    .. math::

        p(\w) = \mathcal{N}(\muw, \sigw)


    We can then find the maximum likelihood solution of the posterior
    distribution :math:`p(\w | \y) \propto p(\y | \w) p(\w)` by solving
    the matrix equation:

    .. math::

        \w = \covw (X^\\top \cov^{-1} \y + \\boldsymbol\sigma^{-2}_\w I \muw)

    Where :math:`\covw` is the covariance matrix of the coefficients:

    .. math::

        \covw^{-1} = (X^\\top \cov^{-1} X + \\boldsymbol\sigma^{-2}_\w I)


    Parameters
    ----------
    lc : `.LightCurve`
        The light curve that needs to be corrected.
    """
    def __init__(self, lc):
        # We don't accept NaN in time or flux.
        if np.any([~np.isfinite(lc.time.value), ~np.isfinite(lc.flux)]):
            raise ValueError('Input light curve has NaNs in time or flux. '
                             'Please remove NaNs before correction '
                             '(e.g. using `lc = lc.remove_nans()`).')
        # We don't accept NaN in flux_err, unless all values are NaN.
        if np.any(~np.isfinite(lc.flux_err)) and not np.all(~np.isfinite(lc.flux_err)):
            raise ValueError('Input light curve has NaNs in `flux_err`. '
                             'Please remove NaNs before correction '
                             '(e.g. using `lc = lc.remove_nans()`).')
        if np.any(lc.flux_err[np.isfinite(lc.flux_err)] <= 0):
            raise ValueError('Input light curve contains flux uncertainties '
                             'smaller than or equal to zero. Please remove '
                             'these (e.g. using `lc = lc[lc.flux_err > 0]`).')
        self.lc = lc

        # The following properties will be set when correct() is called.
        # We're setting them here so they do not throw value errors
        self.design_matrix_collection = None
        self.coefficients = None
        self.corrected_lc = None
        self.model_lc = None
        self.diagnostic_lightcurves = None

    def __repr__(self):
        return 'RegressionCorrector (ID: {})'.format(self.lc.targetid)

    @property
    def dmc(self):
        """Shorthand for self.design_matrix_collection."""
        return self.design_matrix_collection

    def _fit_coefficients(self, cadence_mask=None, prior_mu=None,
                          prior_sigma=None, propagate_errors=False):
        """Fit the linear regression coefficients.

        This function will solve a linear regression with Gaussian priors
        on the coefficients.

        Parameters
        ----------
        cadence_mask : np.ndarray of bool
            Mask, where True indicates a cadence that should be used.

        Returns
        -------
        coefficients : np.ndarray
            The best fit model coefficients to the data.
        """

        # If prior_mu is specified, prior_sigma must be specified
        if not ((prior_mu is None) & (prior_sigma is None)) | \
                    ((prior_mu is not None) & (prior_sigma is not None)):
            raise ValueError("Please specify both `prior_mu` and `prior_sigma`")

        # Default cadence mask
        if cadence_mask is None:
            cadence_mask = np.ones(len(self.lc.flux.value), bool)

        # If flux errors are not all finite numbers, then default to array of ones
        if np.all(~np.isfinite(self.lc.flux_err.value)):
            flux_err = np.ones(cadence_mask.sum())
        else:
            flux_err = self.lc.flux_err.value[cadence_mask]

        # Retrieve the design matrix (X) as a numpy array
        X = self.dmc.X[cadence_mask]
        if isinstance(X, np.ndarray):
            # Compute `X^T cov^-1 X + 1/prior_sigma^2`
            sigma_w_inv = X.T.dot(X / flux_err[:, None]**2)
            # Compute `X^T cov^-1 y + prior_mu/prior_sigma^2`
            B = np.dot(X.T, self.lc.flux.value[cadence_mask] / flux_err**2)

        elif issparse(X):
            sigma_f_inv = csr_matrix(1/flux_err[:, None]**2)
            # Compute `X^T cov^-1 X + 1/prior_sigma^2`
            sigma_w_inv = X.T.dot(X.multiply(sigma_f_inv))
            # Compute `X^T cov^-1 y + prior_mu/prior_sigma^2`
            B = X.T.dot((self.lc.flux[cadence_mask]/flux_err**2))
            sigma_w_inv = sigma_w_inv.toarray()

        if prior_sigma is not None:
            sigma_w_inv = sigma_w_inv + np.diag(1. / prior_sigma**2)
        if prior_sigma is not None:
            B = B + (prior_mu / prior_sigma**2)

        # Solve for weights w
        w = np.linalg.solve(sigma_w_inv, B).T
        if propagate_errors:
            w_err = np.linalg.inv(sigma_w_inv)
        else:
            w_err = np.zeros(len(w)) * np.nan

        return w, w_err

    def correct(self, design_matrix_collection, cadence_mask=None, sigma=5,
                niters=5, propagate_errors=False):
        """Find the best fit correction for the light curve.

        Parameters
        ----------
        design_matrix_collection : `.DesignMatrix` or `.DesignMatrixCollection`
            One or more design matrices.  Each matrix must have a shape of
            (time, regressors). The columns contained in each matrix must be
            known to correlate with additive noise components we want to remove
            from the light curve.
        cadence_mask : np.ndarray of bools (optional)
            Mask, where True indicates a cadence that should be used.
        sigma : int (default 5)
            Standard deviation at which to remove outliers from fitting
        niters : int (default 5)
            Number of iterations to fit and remove outliers
        propagate_errors : bool (default False)
            Whether to propagate the uncertainties from the regression. Default is False.
            Setting to True will increase run time, but will sample from multivariate normal
            distribution of weights.

        Returns
        -------
        `.LightCurve`
            Corrected light curve, with noise removed.
        """
        if not isinstance(design_matrix_collection, DesignMatrixCollection):
            if isinstance(design_matrix_collection, SparseDesignMatrix):
                design_matrix_collection = SparseDesignMatrixCollection([design_matrix_collection])
            elif isinstance(design_matrix_collection, DesignMatrix):
                design_matrix_collection = DesignMatrixCollection([design_matrix_collection])

        # Validate the design matrix. Emits a warning if the matrix has low rank.
        design_matrix_collection.validate()
        self.design_matrix_collection = design_matrix_collection

        if cadence_mask is None:
            self.cadence_mask = np.ones(len(self.lc.time), bool)
        else:
            self.cadence_mask = cadence_mask

        # Create an outlier mask using iterative sigma clipping
        self.outlier_mask = np.zeros_like(self.cadence_mask)
        for count in range(niters):
            tmp_cadence_mask = self.cadence_mask & ~self.outlier_mask
            coefficients, coefficients_err = \
                self._fit_coefficients(cadence_mask=tmp_cadence_mask,
                                       prior_mu=self.dmc.prior_mu,
                                       prior_sigma=self.dmc.prior_sigma,
                                       propagate_errors=propagate_errors)
            model = np.ma.masked_array(data=self.dmc.X.dot(coefficients),
                                       mask=~tmp_cadence_mask)
            model = u.Quantity(model, unit=self.lc.flux.unit)
            residuals = self.lc.flux - model
            self.outlier_mask |= sigma_clip(residuals, sigma=sigma).mask
            log.debug("correct(): iteration {}: clipped {} cadences"
                      "".format(count, self.outlier_mask.sum()))

        self.coefficients = coefficients
        self.coefficients_err = coefficients_err

        model_flux = self.dmc.X.dot(coefficients)
        model_flux -= np.median(model_flux)
        if propagate_errors:
            with warnings.catch_warnings():
                # ignore "RuntimeWarning: covariance is not symmetric positive-semidefinite."
                warnings.simplefilter("ignore", RuntimeWarning)
                samples = np.asarray(
                    [self.dmc.X.dot(np.random.multivariate_normal(coefficients, coefficients_err))
                     for idx in range(100)]).T
            model_err = np.abs(np.percentile(samples, [16, 84], axis=1) - np.median(samples, axis=1)[:, None].T).mean(axis=0)
        else:
            model_err = np.zeros(len(model_flux))
        self.model_lc = LightCurve(time=self.lc.time,
                                   flux=u.Quantity(model_flux, unit=self.lc.flux.unit),
                                   flux_err=u.Quantity(model_err, unit=self.lc.flux.unit))
        self.corrected_lc = self.lc.copy()
        self.corrected_lc.flux = self.lc.flux - self.model_lc.flux
        self.corrected_lc.flux_err = (self.lc.flux_err**2 + model_err**2)**0.5
        self.diagnostic_lightcurves = self._create_diagnostic_lightcurves()
        return self.corrected_lc

    def _create_diagnostic_lightcurves(self):
        """Returns a dictionary containing all diagnostic light curves.

        The dictionary will provide a light curve for each matrix in the
        design matrix collection.
        """
        if self.coefficients is None:
            raise ValueError("you need to call `correct()` first")

        lcs = {}
        for idx, submatrix in enumerate(self.dmc.matrices):
            # What is the index of the first column for the submatrix?
            firstcol_idx = sum([m.shape[1] for m in self.dmc.matrices[:idx]])
            submatrix_coefficients = self.coefficients[firstcol_idx:firstcol_idx+submatrix.shape[1]]
            # submatrix_coefficients_err = self.coefficients_err[firstcol_idx:firstcol_idx+submatrix.shape[1], firstcol_idx:firstcol_idx+submatrix.shape[1]]
            # samples = np.asarray([np.dot(submatrix.values, np.random.multivariate_normal(submatrix_coefficients, submatrix_coefficients_err)) for idx in range(100)]).T
            # model_err = np.abs(np.percentile(samples, [16, 84], axis=1) - np.median(samples, axis=1)[:, None].T).mean(axis=0)
            model_flux = u.Quantity(submatrix.X.dot(submatrix_coefficients), unit=self.lc.flux.unit)
            model_flux_err = u.Quantity(np.zeros(len(model_flux)), unit=self.lc.flux.unit)
            lcs[submatrix.name] = LightCurve(time=self.lc.time, flux=model_flux,
                                             flux_err=model_flux_err, label=submatrix.name)
        return lcs

    def _diagnostic_plot(self):
        """Produce diagnostic plots to assess the effectiveness of the correction.

        Note: We need a hidden function so that other correctors can alter the plot.
        """
        if not hasattr(self, 'corrected_lc'):
            raise ValueError('Please call the `correct()` method before trying to diagnose.')

        with plt.style.context(MPLSTYLE):
            _, axs = plt.subplots(2, figsize=(10, 6), sharex=True)
            ax = axs[0]
            self.lc.plot(ax=ax, normalize=False, label='original', alpha=0.4)
            for key in self.diagnostic_lightcurves.keys():
                (self.diagnostic_lightcurves[key] - np.median(self.diagnostic_lightcurves[key].flux) + np.median(self.lc.flux)).plot(ax=ax)
            ax.set_xlabel('')
            ax = axs[1]
            self.lc.plot(ax=ax, normalize=False, alpha=0.2, label='original')
            self.corrected_lc[self.outlier_mask].scatter(
                                            normalize=False, c='r', marker='x',
                                            s=10, label='outlier_mask', ax=ax)
            self.corrected_lc[~self.cadence_mask].scatter(
                                            normalize=False, c='dodgerblue',
                                            marker='x', s=10, label='~cadence_mask',
                                            ax=ax)
            self.corrected_lc.plot(normalize=False, label='corrected', ax=ax, c='k')
        return axs

    def diagnose(self):
        """Returns diagnostic plots to assess the most recent call to `correct()`.

        If `correct()` has not yet been called, a ``ValueError`` will be raised.

        Returns
        -------
        `~matplotlib.axes.Axes`
            The matplotlib axes object.
        """
        return self._diagnostic_plot()

    def diagnose_priors(self):
        """Returns a diagnostic plot visualizing how the best-fit coefficients
        compare against the priors.

        The method will show the results obtained during the most recent call
        to `correct()`.  If `correct()` has not yet been called, a
        ``ValueError`` will be raised.

        Returns
        -------
        `~matplotlib.axes.Axes`
            The matplotlib axes object.
        """
        if not hasattr(self, 'corrected_lc'):
            raise ValueError('Please call the `correct()` method before trying to diagnose.')

        names = [dm.name for dm in self.dmc]
        with plt.style.context(MPLSTYLE):
            _, axs = plt.subplots(1, len(names), figsize=(len(names)*4, 4),
                                  sharey=True)
            if not hasattr(axs, '__iter__'):
                axs = [axs]
            for idx, ax, X in zip(range(len(names)), axs, self.dmc):
                X.plot_priors(ax=ax)
                firstcol_idx = sum([m.shape[1] for m in self.dmc.matrices[:idx]])
                submatrix_coefficients = self.coefficients[firstcol_idx:firstcol_idx+X.shape[1]]
                [ax.axvline(s, color='red', zorder=-1) for s in submatrix_coefficients]
        return axs



"""Defines the Seismology class."""
import logging
import warnings

import numpy as np
from matplotlib import pyplot as plt
from scipy.signal import find_peaks

from astropy import units as u
from astropy.units import cds

from .. import MPLSTYLE
from . import utils, stellar_estimators
from ..periodogram import SNRPeriodogram
from ..utils import LightkurveWarning, validate_method
from .utils import  SeismologyQuantity

# Import the optional Bokeh dependency required by ``interact_echelle```,
# or print a friendly error otherwise.
try:
    import bokeh  # Import bokeh first so we get an ImportError we can catch
    from bokeh.io import show, output_notebook
    from bokeh.plotting import figure
    from bokeh.models import LogColorMapper, Slider, RangeSlider, Button
    from bokeh.layouts import layout, Spacer
except:
    # Nice error will be raised when ``interact_echelle``` is called.
    pass

log = logging.getLogger(__name__)

__all__ = ['Seismology']


class Seismology(object):
    """Enables astroseismic quantities to be estimated from periodograms.

    This class provides easy access to methods to estimate numax, deltanu, radius,
    mass, and logg, and stores them on its tray for easy diagnostic plotting.

    Examples
    --------
    Download the TESS light curve for HIP 116158:

        >>> import lightkurve as lk
        >>> lc = lk.search_lightcurve("HIP 116158", sector=2).download()  # doctest: +SKIP
        >>> lc = lc.normalize().remove_nans().remove_outliers()  # doctest: +SKIP

    Create a Lomb-Scargle periodogram:

        >>> pg = lc.to_periodogram(normalization='psd', minimum_frequency=100, maximum_frequency=800)  # doctest: +SKIP

    Create a Seismology object and use it to estimate parameters:

        >>> seismology = pg.flatten().to_seismology()  # doctest: +SKIP
        >>> seismology.estimate_numax()  # doctest: +SKIP
        numax: 415.00 uHz (method: ACF2D)
        >>> seismology.estimate_deltanu()  # doctest: +SKIP
        deltanu: 28.78 uHz (method: ACF2D)
        >>> seismology.estimate_radius(teff=5080)  # doctest: +SKIP
        radius: 2.78 solRad (method: Uncorrected Scaling Relations)

    Parameters
    ----------
    periodogram : `~lightkurve.periodogram.Periodogram` object
        Periodogram to be analyzed. Must be background-corrected,
        e.g. using `periodogram.flatten()`.
    """
    def __init__(self, periodogram):
        if not isinstance(periodogram, SNRPeriodogram):
            warnings.warn("Seismology received a periodogram which does not appear "
                          "to have been background-corrected. Please consider calling "
                          "`periodogram.flatten()` prior to extracting seismological parameters.",
                          LightkurveWarning)
        self.periodogram = periodogram

    def __repr__(self):
        attrs = np.asarray(['numax', 'deltanu', 'mass', 'radius', 'logg'])
        tray = np.asarray([hasattr(self, attr) for attr in attrs])
        if tray.sum() == 0:
            tray_str = " - no values have been computed so far."
        else:
            tray_str = " - computed values:\n * " + "\n * ".join([getattr(self, attr).__repr__() for attr in attrs[tray]])
        return 'Seismology(ID: {}){}'.format(self.periodogram.label, tray_str)

    @staticmethod
    def from_lightcurve(lc, **kwargs):
        """Returns a `Seismology` object given a `~lightkurve.lightcurve.LightCurve` object."""
        log.info("Building a Seismology object directly from a light curve "
                 "uses default periodogram parameters. For further tuneability, "
                 "create a periodogram object first, using `to_periodogram`.")
        return Seismology(periodogram=lc.normalize().remove_nans().fill_gaps().to_periodogram(**kwargs).flatten())

    def _validate_numax(self, numax):
        """Raises exception if `numax` is None and `self.numax` is not set."""
        if numax is None:
            try:
                return self.numax
            except AttributeError:
                raise AttributeError("You need to call `Seismology.estimate_numax()` first.")
        return numax

    def _validate_deltanu(self, deltanu):
        """Raises exception if `deltanu` is None and `self.deltanu` is not set."""
        if deltanu is None:
            try:
                return self.deltanu
            except AttributeError:
                raise AttributeError("You need to call `Seismology.estimate_deltanu()` first.")
        return deltanu


    def _clean_echelle(self, deltanu=None, numax=None,
                         minimum_frequency=None, maximum_frequency=None,
                         smooth_filter_width=.1, scale='linear'):
        """Takes input seismology object and creates the necessary arrays for an echelle
        diagram. Validates all the inputs.

        Parameters
        ----------
        deltanu : float
            Value for the large frequency separation of the seismic mode
            frequencies in the periodogram. Assumed to have the same units as
            the frequencies, unless given an Astropy unit.
            Is assumed to be in the same units as frequency if not given a unit.
        numax : float
            Value for the frequency of maximum oscillation. If a numax is
            passed, a suitable range one FWHM of the mode envelope either side
            of the will be shown. This is overwritten by custom frequency ranges.
            Is assumed to be in the same units as frequency if not given a unit.
        minimum_frequency : float
            The minimum frequency at which to display the echelle
            Is assumed to be in the same units as frequency if not given a unit.
        maximum_frequency : float
            The maximum frequency at which to display the echelle.
            Is assumed to be in the same units as frequency if not given a unit.
        smooth_filter_width : float
            If given a value, will smooth periodogram used to plot the echelle
            diagram using the periodogram.smooth(method='boxkernel') method with
            a filter width of `smooth_filter_width`. This helps visualise the
            echelle diagram. Is assumed to be in the same units as the
            periodogram frequency.
        ax : `~matplotlib.axes.Axes`
            A matplotlib axes object to plot into. If no axes is provided,
            a new one will be created.
        scale: str
            Set z axis to be "linear" or "log". Default is linear.

        Returns
        -------
        ep : np.ndarray
            Echelle diagram power
        x_f : np.ndarray
            frequencies for X axis
        y_f : np.ndarray
            frequencies for Y axis
        """
        if (minimum_frequency is None) & (maximum_frequency is None):
            numax = self._validate_numax(numax)
        deltanu = self._validate_deltanu(deltanu)

        if (not hasattr(numax, 'unit')) & (numax is not None):
            numax = numax * self.periodogram.frequency.unit
        if (not hasattr(deltanu, 'unit')) & (deltanu is not None):
            deltanu = deltanu * self.periodogram.frequency.unit

        if smooth_filter_width:
            pgsmooth = self.periodogram.smooth(filter_width=smooth_filter_width)
            freq = pgsmooth.frequency  # Makes code below more readable below
            power = pgsmooth.power     # Makes code below more readable below
        else:
            freq = self.periodogram.frequency  # Makes code below more readable
            power = self.periodogram.power     # Makes code below more readable

        fmin = freq[0]
        fmax = freq[-1]

        # Check for any superfluous input
        if (numax is not None) & (any([a is not None for a in [minimum_frequency, maximum_frequency]])):
            warnings.warn("You have passed both a numax and a frequency limit. "
                          "The frequency limit will override the numax input.",
                          LightkurveWarning)

        # Ensure input numax is in the correct units (if there is one)
        if numax is not None:
            numax = u.Quantity(numax, freq.unit).value
            if numax > freq[-1].value:
                raise ValueError("You can't pass in a numax outside the"
                                "frequency range of the periodogram.")

            fwhm = utils.get_fwhm(self.periodogram, numax)

            fmin = numax - 2*fwhm
            if fmin < freq[0].value:
                fmin = freq[0].value

            fmax = numax + 2*fwhm
            if fmax > freq[-1].value:
                fmax = freq[-1].value

        # Set limits and set them in the right units
        if minimum_frequency is not None:
            fmin =  u.Quantity(minimum_frequency, freq.unit).value
            if fmin > freq[-1].value:
                raise ValueError("You can't pass in a limit outside the "
                                 "frequency range of the periodogram.")

        if maximum_frequency is not None:
            fmax = u.Quantity(maximum_frequency, freq.unit).value
            if fmax > freq[-1].value:
                raise ValueError("You can't pass in a limit outside the "
                                 "frequency range of the periodogram.")

        # Make sure fmin and fmax are Quantities or code below will break
        fmin = u.Quantity(fmin, freq.unit)
        fmax = u.Quantity(fmax, freq.unit)

        # Add on 1x deltanu so we don't miss off any important range due to rounding
        if fmax < freq[-1] - 1.5*deltanu:
            fmax += deltanu

        fs = np.median(np.diff(freq))
        x0 = int(freq[0] / fs)

        ff = freq[int(fmin/fs)-x0:int(fmax/fs)-x0] # Selected frequency range
        pp = power[int(fmin/fs)-x0:int(fmax/fs)-x0] # Power range

        # Reshape the power into n_rows of n_columns
        #  When modulus ~ zero, deltanu divides into frequency without remainder
        mod_zeros = find_peaks( -1.0*(ff % deltanu) )[0]

        # The bottom left corner of the plot is the lowest frequency that
        # divides into deltanu with almost zero remainder
        start = mod_zeros[0]

        # The top left corner of the plot is the highest frequency that
        # divides into deltanu with almost zero remainder.  This index is the
        # approximate end, because we fix an integer number of rows and columns
        approx_end = mod_zeros[-1]

        # The number of rows is the number of times you can partition your
        #  frequency range into chunks of size deltanu, start and ending at
        #  frequencies that divide nearly evenly into deltanu
        n_rows = len(mod_zeros) - 1

        # The number of columns is the total number of frequency points divided
        #  by the number of rows, floor divided to the nearest integer value
        n_columns =  int( (approx_end - start) / n_rows )

        # The exact end point is therefore the ncolumns*nrows away from the start
        end = start + n_columns*n_rows

        ep = np.reshape(pp[start : end], (n_rows, n_columns))

        if scale=='log':
            ep = np.log10(ep)

        # Reshape the freq into n_rowss of n_columnss & create arays
        ef = np.reshape(ff[start : end], (n_rows, n_columns))
        x_f = ((ef[0,:]-ef[0,0]) % deltanu)
        y_f = (ef[:,0])
        return ep, x_f, y_f

    def plot_echelle(self, deltanu=None, numax=None, minimum_frequency=None,
                     maximum_frequency=None, smooth_filter_width=.1,
                     scale='linear', ax=None, cmap='Blues'):
        """Plots an echelle diagram of the periodogram by stacking the
        periodogram in slices of deltanu.

        Modes of equal radial degree should appear approximately vertically aligned.
        If no structure is present, you are likely dealing with a faulty deltanu
        value or a low signal to noise case.

        This method is adapted from work by Daniel Hey & Guy Davies.

        Parameters
        ----------
        deltanu : float
            Value for the large frequency separation of the seismic mode
            frequencies in the periodogram. Assumed to have the same units as
            the frequencies, unless given an Astropy unit.
            Is assumed to be in the same units as frequency if not given a unit.
        numax : float
            Value for the frequency of maximum oscillation. If a numax is
            passed, a suitable range one FWHM of the mode envelope either side
            of the will be shown. This is overwritten by custom frequency ranges.
            Is assumed to be in the same units as frequency if not given a unit.
        minimum_frequency : float
            The minimum frequency at which to display the echelle
            Is assumed to be in the same units as frequency if not given a unit.
        maximum_frequency : float
            The maximum frequency at which to display the echelle.
            Is assumed to be in the same units as frequency if not given a unit.
        smooth_filter_width : float
            If given a value, will smooth periodogram used to plot the echelle
            diagram using the periodogram.smooth(method='boxkernel') method with
            a filter width of `smooth_filter_width`. This helps visualise the
            echelle diagram. Is assumed to be in the same units as the
            periodogram frequency.
        scale: str
            Set z axis to be "linear" or "log". Default is linear.
        cmap : str
            The name of the matplotlib colourmap to use in the echelle diagram.

        Returns
        -------
        ax : `~matplotlib.axes.Axes`
            The matplotlib axes object.
        """
        if (minimum_frequency is None) & (maximum_frequency is None):
            numax = self._validate_numax(numax)
        deltanu = self._validate_deltanu(deltanu)

        if (not hasattr(numax, 'unit')) & (numax is not None):
            numax = numax * self.periodogram.frequency.unit
        if (not hasattr(deltanu, 'unit')) & (deltanu is not None):
            deltanu = deltanu * self.periodogram.frequency.unit

        ep, x_f, y_f = self._clean_echelle(numax=numax, deltanu=deltanu,
                                           minimum_frequency=minimum_frequency,
                                           maximum_frequency=maximum_frequency,
                                           smooth_filter_width=smooth_filter_width)

        # Plot the echelle diagram
        with plt.style.context(MPLSTYLE):
            if ax is None:
                _, ax = plt.subplots()
            extent = (x_f[0].value, x_f[-1].value, y_f[0].value, y_f[-1].value)
            figsize = plt.rcParams['figure.figsize']
            a = figsize[1] / figsize[0]
            b = (extent[3] - extent[2]) / (extent[1] - extent[0])
            vmin = np.nanpercentile(ep.value, 1)
            vmax = np.nanpercentile(ep.value, 99)

            im = ax.imshow(ep.value, cmap=cmap, aspect=a/b, origin='lower',
                           extent=extent, vmin=vmin, vmax=vmax)

            cbar = plt.colorbar(im, ax=ax, extend='both', pad=.01)

            if isinstance(self.periodogram, SNRPeriodogram):
                ylabel = 'Signal to Noise Ratio (SNR)'
            elif self.periodogram.power.unit == cds.ppm:
                ylabel = "Amplitude [{}]".format(self.periodogram.power.unit.to_string('latex'))
            else:
                ylabel = "Power Spectral Density [{}]".format(self.periodogram.power.unit.to_string('latex'))

            if scale == 'log':
                ylabel = 'log10('+ylabel+')'

            cbar.set_label(ylabel)
            ax.set_xlabel(r'Frequency mod. {:.2f}'.format(deltanu))
            ax.set_ylabel(r'Frequency [{}]'.format(self.periodogram.frequency.unit.to_string('latex')))
            ax.set_title('Echelle diagram for {}'.format(self.periodogram.label))

        return ax

    def _make_echelle_elements(self, deltanu, cmap='viridis',
        minimum_frequency=None, maximum_frequency=None, smooth_filter_width=.1,
        scale='linear', plot_width=490, plot_height=340, title='Echelle'):
        """Helper function to make the elements of the echelle diagram for bokeh plotting.
        """
        if not hasattr(deltanu, 'unit'):
            deltanu = deltanu * self.periodogram.frequency.unit

        if smooth_filter_width:
            pgsmooth = self.periodogram.smooth(filter_width=smooth_filter_width)
            freq = pgsmooth.frequency  # Makes code below more readable below
        else:
            freq = self.periodogram.frequency  # Makes code below more readable

        ep, x_f, y_f = self._clean_echelle(deltanu=deltanu,
                                           minimum_frequency=minimum_frequency,
                                           maximum_frequency=maximum_frequency,
                                           smooth_filter_width=smooth_filter_width,
                                           scale=scale)

        fig = figure(plot_width=plot_width, plot_height=plot_height,
                     x_range=(0, 1), y_range=(y_f[0].value, y_f[-1].value),
                     title=title, tools='pan,box_zoom,reset',
                     toolbar_location="above",
                     border_fill_color="white")

        fig.yaxis.axis_label = r'Frequency [{}]'.format(freq.unit.to_string())
        fig.xaxis.axis_label = r'Frequency / {:.3f} Mod. 1'.format(deltanu)

        lo, hi = np.nanpercentile(ep.value, [0.1, 99.9])
        vlo, vhi = 0.3 * lo, 1.7 * hi
        vstep = (lo - hi)/500
        color_mapper = LogColorMapper(palette="RdYlGn10", low=lo, high=hi)

        fig.image(image=[ep.value], x=0, y=y_f[0].value,
                  dw=1, dh=y_f[-1].value,
                  color_mapper=color_mapper, name='img')

        stretch_slider = RangeSlider(start=vlo,
                                     end=vhi,
                                     step=vstep,
                                     title='',
                                     value=(lo, hi),
                                     orientation='vertical',
                                     width=10,
                                     height=230,
                                     direction='rtl',
                                     show_value=False,
                                     sizing_mode='fixed',
                                     name='stretch')

        def stretch_change_callback(attr, old, new):
            """TPF stretch slider callback."""
            fig.select('img')[0].glyph.color_mapper.high = new[1]
            fig.select('img')[0].glyph.color_mapper.low = new[0]

        stretch_slider.on_change('value', stretch_change_callback)
        return fig, stretch_slider


    def interact_echelle(self, notebook_url="localhost:8888", **kwargs):
        """Display an interactive Jupyter notebook widget showing an Echelle diagram.

        This feature only works inside an active Jupyter Notebook, and
        requires an optional dependency, ``bokeh`` (v1.0 or later).
        This dependency can be installed using e.g. `conda install bokeh`.

        Parameters
        ----------
        notebook_url : str
            Location of the Jupyter notebook page (default: "localhost:8888")
            When showing Bokeh applications, the Bokeh server must be
            explicitly configured to allow connections originating from
            different URLs. This parameter defaults to the standard notebook
            host and port. If you are running on a different location, you
            will need to supply this value for the application to display
            properly. If no protocol is supplied in the URL, e.g. if it is
            of the form "localhost:8888", then "http" will be used.
        """
        try:
            import bokeh
            if bokeh.__version__[0] == '0':
                warnings.warn("interact() requires Bokeh version 1.0 or later", LightkurveWarning)
        except ImportError:
            log.error("The interact() tool requires the `bokeh` Python package; "
                      "you can install bokeh using e.g. `conda install bokeh`.")
            return None

        maximum_frequency = kwargs.pop('maximum_frequency', self.periodogram.frequency.max().value)
        minimum_frequency = kwargs.pop('minimum_frequency', self.periodogram.frequency. min().value)

        if not hasattr(self, 'deltanu'):
            dnu = SeismologyQuantity(quantity=self.periodogram.frequency.max()/30,
                                     name='deltanu', method='echelle')
        else:
            dnu = self.deltanu

        def create_interact_ui(doc):
            fig_tpf, stretch_slider = self._make_echelle_elements(dnu,
                                              maximum_frequency=maximum_frequency,
                                              minimum_frequency=minimum_frequency,
                                              **kwargs)
            maxdnu = self.periodogram.frequency.max().value/5
            # Interactive slider widgets
            dnu_slider = Slider(start=0.01,
                                end=maxdnu,
                                value=dnu.value,
                                step=0.01,
                                title="Delta Nu",
                                width=290)
            r_button = Button(label=">", button_type="default", width=30)
            l_button = Button(label="<", button_type="default", width=30)
            rr_button = Button(label=">>", button_type="default", width=30)
            ll_button = Button(label="<<", button_type="default", width=30)

            def update(attr, old, new):
                """Callback to take action when dnu slider changes"""
                dnu = SeismologyQuantity(quantity=dnu_slider.value*u.microhertz,
                                         name='deltanu', method='echelle')
                ep, _, _ = self._clean_echelle(deltanu=dnu,
                                               minimum_frequency=minimum_frequency,
                                               maximum_frequency=maximum_frequency,
                                               **kwargs)
                fig_tpf.select('img')[0].data_source.data['image'] = [ep.value]
                fig_tpf.xaxis.axis_label = r'Frequency / {:.3f} Mod. 1'.format(dnu)

            def go_right_by_one_small():
                """Step forward in time by a single cadence"""
                existing_value = dnu_slider.value
                if existing_value < maxdnu:
                    dnu_slider.value = existing_value + 0.002

            def go_left_by_one_small():
                """Step back in time by a single cadence"""
                existing_value = dnu_slider.value
                if existing_value > 0:
                    dnu_slider.value = existing_value - 0.002

            def go_right_by_one():
                """Step forward in time by a single cadence"""
                existing_value = dnu_slider.value
                if existing_value < maxdnu:
                    dnu_slider.value = existing_value + 0.01

            def go_left_by_one():
                """Step back in time by a single cadence"""
                existing_value = dnu_slider.value
                if existing_value > 0:
                    dnu_slider.value = existing_value - 0.01

            dnu_slider.on_change('value', update)
            r_button.on_click(go_right_by_one_small)
            l_button.on_click(go_left_by_one_small)
            rr_button.on_click(go_right_by_one)
            ll_button.on_click(go_left_by_one)

            widgets_and_figures = layout([fig_tpf, [Spacer(height=20), stretch_slider]],
                                         [ll_button, Spacer(width=30), l_button,
                                          Spacer(width=25), dnu_slider, Spacer(width=30),
                                          r_button, Spacer(width=23), rr_button])
            doc.add_root(widgets_and_figures)

        output_notebook(verbose=False, hide_banner=True)
        return show(create_interact_ui, notebook_url=notebook_url)

    def estimate_numax(self, method="acf2d", **kwargs):
        """Returns the frequency of the peak of the seismic oscillation modes envelope.

        At present, the only method supported is based on using a
        2D autocorrelation function (ACF2D).  This method is implemented by the
        `~lightkurve.seismology.estimate_numax_acf2d` function which accepts
        the parameters `numaxs`, `window_width`, and `spacing`.
        For details and literature references, please read the detailed
        docstring of this function by typing ``lightkurve.seismology.estimate_numax_acf2d?``
        in a Python terminal or notebook.

        Parameters
        ----------
        method : str
            Method to use. Only ``"acf2d"`` is supported at this time.

        Returns
        -------
        numax : `~lightkurve.seismology.SeismologyQuantity`
            Numax of the periodogram, including details on the units and method.
        """
        method = validate_method(method, supported_methods=["acf2d"])
        if method == "acf2d":
            from .numax_estimators import estimate_numax_acf2d
            result = estimate_numax_acf2d(self.periodogram, **kwargs)
        self.numax = result
        return result

    def diagnose_numax(self, numax=None):
        """Create diagnostic plots showing how numax was estimated."""
        numax = self._validate_numax(numax)
        return numax.diagnostics_plot_method(numax, self.periodogram)

    def estimate_deltanu(self, method='acf2d', numax=None):
        """Returns the average value of the large frequency spacing, DeltaNu,
        of the seismic oscillations of the target.

        At present, the only method supported is based on using an
        autocorrelation function (ACF2D).  This method is implemented by the
        `~lightkurve.seismology.estimate_deltanu_acf2d` function which requires
        the parameter `numax`. For details and literature references, please
        read the detailed docstring of this function by typing
        ``lightkurve.seismology.estimate_deltanu_acf2d?`` in a Python terminal or notebook.

        Parameters
        ----------
        method : str
            Method to use. Only ``"acf2d"`` is supported at this time.

        Returns
        -------
        deltanu : `~lightkurve.seismology.SeismologyQuantity`
            DeltaNu of the periodogram, including details on the units and method.
        """
        method = validate_method(method, supported_methods=["acf2d"])
        numax = self._validate_numax(numax)
        if method == "acf2d":
            from .deltanu_estimators import estimate_deltanu_acf2d
            result = estimate_deltanu_acf2d(self.periodogram, numax=numax)
        self.deltanu = result
        return result

    def diagnose_deltanu(self, deltanu=None):
        """Create diagnostic plots showing how numax was estimated."""
        deltanu = self._validate_deltanu(deltanu)
        return deltanu.diagnostics_plot_method(deltanu, self.periodogram)

    def estimate_radius(self, teff=None, numax=None, deltanu=None):
        """Returns a stellar radius estimate based on the scaling relations.

        The two global observable seismic parameters, numax and deltanu, along with
        temperature, scale with fundamental stellar properties (Brown et al. 1991;
        Kjeldsen & Bedding 1995). These scaling relations can be rearranged to
        calculate a stellar radius as

        R = Rsol * (numax/numax_sol)(deltanu/deltanusol)^-2(Teff/Teffsol)^0.5

        where R is the radius and Teff is the effective temperature, and the suffix
        'sol' indicates a solar value. In this method we use the solar values for
        numax and deltanu as given in Huber et al. (2011) and for Teff as given in
        Prsa et al. (2016).

        This code structure borrows from work done in Bellinger et al. (2019), which
        also functions as an accessible explanation of seismic scaling relations.

        If no value of effective temperature is given, this function will check the
        meta data of the `Periodogram` object used to create the `Seismology` object.
        These data will often contain an effective tempearture from the Kepler Input
        Catalogue (KIC, https://ui.adsabs.harvard.edu/abs/2011AJ....142..112B/abstract),
        or from the EPIC or TIC for K2 and TESS respectively. The temperature values in these
        catalogues are estimated using photometry, and so have large associated uncertainties
        (roughly 200 K, see KIC). For more better results, spectroscopic measurements of
        temperature are often more precise.

        NOTE: These scaling relations are scaled to the Sun, and therefore do not
        always produce an entirely accurate result for more evolved stars.

        Parameters
        ----------
        numax : float
            The frequency of maximum power of the seismic mode envelope. If not
            given an astropy unit, assumed to be in units of microhertz.
        deltanu : float
            The frequency spacing between two consecutive overtones of equal radial
            degree. If not given an astropy unit, assumed to be in units of
            microhertz.
        teff : float
            The effective temperature of the star. In units of Kelvin.
        numax_err : float
            Error on numax. Assumed to be same units as numax
        deltanu_err : float
            Error on deltanu. Assumed to be same units as deltanu
        teff_err : float
            Error on Teff. Assumed to be same units as Teff.

        Returns
        -------
        radius : `~lightkurve.seismology.SeismologyQuantity`
            Stellar radius estimate.
        """
        numax = self._validate_numax(numax)
        deltanu = self._validate_deltanu(deltanu)
        if teff is None:
            teff = self.periodogram.meta.get('TEFF')
            if teff is None:
                raise ValueError("You must provide an effective temperature argument (`teff`) to `estimate_radius`,"
                        "because the Periodogram object does not contain it in its meta data (i.e. `pg.meta['TEFF']` is missing")
            else:
                log.info("Using value for effective temperature from the Kepler Input Catalogue."
                        "These temperatue values may sometimes differ significantly from modern estimates.")
                pass
        else:
            pass

        result = stellar_estimators.estimate_radius(numax, deltanu, teff)
        self.radius = result
        return result

    def estimate_mass(self, teff=None, numax=None, deltanu=None):
        """Calculates mass using the asteroseismic scaling relations.

        The two global observable seismic parameters, numax and deltanu, along with
        temperature, scale with fundamental stellar properties (Brown et al. 1991;
        Kjeldsen & Bedding 1995). These scaling relations can be rearranged to
        calculate a stellar mass as

        M = Msol * (numax/numax_sol)^3(deltanu/deltanusol)^-4(Teff/Teffsol)^1.5

        where M is the mass and Teff is the effective temperature, and the suffix
        'sol' indicates a solar value. In this method we use the solar values for
        numax and deltanu as given in Huber et al. (2011) and for Teff as given in
        Prsa et al. (2016).

        This code structure borrows from work done in Bellinger et al. (2019), which
        also functions as an accessible explanation of seismic scaling relations.

        If no value of effective temperature is given, this function will check the
        meta data of the `Periodogram` object used to create the `Seismology` object.
        These data will often contain an effective tempearture from the Kepler Input
        Catalogue (KIC, https://ui.adsabs.harvard.edu/abs/2011AJ....142..112B/abstract),
        or from the EPIC or TIC for K2 and TESS respectively. The temperature values in these
        catalogues are estimated using photometry, and so have large associated uncertainties
        (roughly 200 K, see KIC). For more better results, spectroscopic measurements of
        temperature are often more precise.

        NOTE: These scaling relations are scaled to the Sun, and therefore do not
        always produce an entirely accurate result for more evolved stars.

        Parameters
        ----------
        numax : float
            The frequency of maximum power of the seismic mode envelope. If not
            given an astropy unit, assumed to be in units of microhertz.
        deltanu : float
            The frequency spacing between two consecutive overtones of equal radial
            degree. If not given an astropy unit, assumed to be in units of
            microhertz.
        teff : float
            The effective temperature of the star. In units of Kelvin.
        numax_err : float
            Error on numax. Assumed to be same units as numax
        deltanu_err : float
            Error on deltanu. Assumed to be same units as deltanu
        teff_err : float
            Error on Teff. Assumed to be same units as Teff.

        Returns
        -------
        mass : `~lightkurve.seismology.SeismologyQuantity`
            Stellar mass estimate.
        """
        numax = self._validate_numax(numax)
        deltanu = self._validate_deltanu(deltanu)
        if teff is None:
            teff = self.periodogram.meta.get('TEFF')
            if teff is None:
                raise ValueError("You must provide an effective temperature argument (`teff`) to `estimate_radius`,"
                        "because the Periodogram object does not contain it in its meta data (i.e. `pg.meta['TEFF']` is missing")
            else:
                log.info("Using value for effective temperature from the Kepler Input Catalogue."
                        "These temperatue values may sometimes differ significantly from modern estimates.")
                pass
        else:
            pass

        result = stellar_estimators.estimate_mass(numax, deltanu, teff)
        self.mass = result
        return result

    def estimate_logg(self, teff=None, numax=None):
        """Calculates the log of the surface gravity using the asteroseismic scaling
        relations.

        The two global observable seismic parameters, numax and deltanu, along with
        temperature, scale with fundamental stellar properties (Brown et al. 1991;
        Kjeldsen & Bedding 1995). These scaling relations can be rearranged to
        calculate a stellar surface gravity as

            g = gsol * (numax/numax_sol)(Teff/Teffsol)^0.5

        where g is the surface gravity and Teff is the effective temperature,
        and the suffix 'sol' indicates a solar value. In this method we use the
        solar values for numax as given in Huber et al. (2011) and for Teff as given
        in Prsa et al. (2016). The solar surface gravity is calcluated from the
        astropy constants for solar mass and radius and does not have an error.

        The solar surface gravity is returned as log10(g) with units in dex, as is
        common in the astrophysics literature.

        This code structure borrows from work done in Bellinger et al. (2019), which
        also functions as an accessible explanation of seismic scaling relations.

        If no value of effective temperature is given, this function will check the
        meta data of the `Periodogram` object used to create the `Seismology` object.
        These data will often contain an effective tempearture from the Kepler Input
        Catalogue (KIC, https://ui.adsabs.harvard.edu/abs/2011AJ....142..112B/abstract),
        or from the EPIC or TIC for K2 and TESS respectively. The temperature values in these
        catalogues are estimated using photometry, and so have large associated uncertainties
        (roughly 200 K, see KIC). For more better results, spectroscopic measurements of
        temperature are often more precise.

        NOTE: These scaling relations are scaled to the Sun, and therefore do not
        always produce an entirely accurate result for more evolved stars.

        Parameters
        ----------
        numax : float
            The frequency of maximum power of the seismic mode envelope. If not
            given an astropy unit, assumed to be in units of microhertz.
        teff : float
            The effective temperature of the star. In units of Kelvin.
        numax_err : float
            Error on numax. Assumed to be same units as numax
        teff_err : float
            Error on teff. Assumed to be same units as teff.

        Returns
        -------
        logg : `~lightkurve.seismology.SeismologyQuantity`
            Stellar surface gravity estimate.
        """
        numax = self._validate_numax(numax)
        if teff is None:
            teff = self.periodogram.meta.get('TEFF')
            if teff is None:
                raise ValueError("You must provide an effective temperature argument (`teff`) to `estimate_radius`,"
                        "because the Periodogram object does not contain it in its meta data (i.e. `pg.meta['TEFF']` is missing")
            else:
                log.info("Using value for effective temperature from the Kepler Input Catalogue."
                        "These temperatue values may sometimes differ significantly from modern estimates.")
                pass
        else:
            pass
        result = stellar_estimators.estimate_logg(numax, teff)
        self.logg = result
        return result



"""Tests the features of the lightkurve.interact_bls module."""
import pytest

from astropy.timeseries import BoxLeastSquares
import astropy.units as u
import numpy as np

from ..lightcurve import KeplerLightCurve, TessLightCurve
from .test_lightcurve import KEPLER10, TESS_SIM

bad_optional_imports = False
try:
    import bokeh
except:
    bad_optional_imports = True


@pytest.mark.remote_data
@pytest.mark.skipif(bad_optional_imports,
                    reason="requires bokeh and astropy.stats.bls")
def test_malformed_notebook_url():
    """Test if malformed notebook_urls raise proper exceptions."""
    lc = KeplerLightCurve.read(KEPLER10)
    lc = lc.normalize().remove_nans().flatten()
    with pytest.raises(ValueError) as exc:
        lc.interact_bls(notebook_url='')
    assert('Empty host value' in exc.value.args[0])
    with pytest.raises(AttributeError) as exc:
        lc.interact_bls(notebook_url=None)
    assert('object has no attribute' in exc.value.args[0])

@pytest.mark.remote_data
@pytest.mark.skipif(bad_optional_imports,
                    reason="requires bokeh and astropy.stats.bls")
def test_graceful_exit_outside_notebook():
    """Test if running interact outside of a notebook does fails gracefully."""
    lc = KeplerLightCurve.read(KEPLER10)
    lc = lc.normalize().remove_nans().flatten()
    result = lc.interact_bls()
    assert(result is None)

@pytest.mark.remote_data
@pytest.mark.skipif(bad_optional_imports,
                    reason="requires bokeh and astropy.stats.bls")
def test_helper_functions():
    """Can we use all the functions in interact_bls?"""
    from ..interact_bls import (prepare_bls_datasource,
                        prepare_folded_datasource, prepare_lightcurve_datasource)
    from ..interact_bls import (make_bls_figure_elements,
                                        make_folded_figure_elements,
                                        make_lightcurve_figure_elements)
    from ..interact_bls import (prepare_bls_help_source,
                                        prepare_f_help_source,
                                        prepare_lc_help_source)
    lc = KeplerLightCurve.read(KEPLER10)
    lc = lc.normalize().remove_nans().flatten()
    lc_source = prepare_lightcurve_datasource(lc)
    f_source = prepare_folded_datasource(lc.fold(1))
    model = BoxLeastSquares(lc.time, lc.flux)
    result = model.power([1,2,3], 0.3)
    bls_source = prepare_bls_datasource(result, 0)

    lc_help = prepare_lc_help_source(lc)
    f_help = prepare_f_help_source(lc.fold(1))
    bls_help = prepare_bls_help_source(bls_source, 1)

    make_lightcurve_figure_elements(lc, lc, lc_source, lc_source, lc_help)
    make_folded_figure_elements(lc.fold(1), lc.fold(1), f_source, f_source, f_help)
    make_bls_figure_elements(result, bls_source, bls_help)


@pytest.mark.remote_data
def test_preprocess_lc():
    '''Test to ensure the lightcurve is pre-processed before applying BLS for correctness and consistent output'''
    from ..interact_bls import _preprocess_lc_for_bls
    lc = KeplerLightCurve.read(KEPLER10)
    assert np.isnan(lc.flux).any()  # ensure the test data has nan in flux

    clean = _preprocess_lc_for_bls(lc)
    assert not np.isnan(clean.flux).any()   # ensure processed lc has no nan
    assert clean.meta.get('NORMALIZED', False)
    assert clean.flux.unit == u.dimensionless_unscaled

    # case the lc is normalized, but in other units
    lc = lc.normalize(unit='percent')
    clean = _preprocess_lc_for_bls(lc)
    assert not np.isnan(clean.flux).any()   # ensure processed lc has no nan
    assert clean.meta.get('NORMALIZED', False)
    assert clean.flux.unit == u.dimensionless_unscaled


@pytest.mark.remote_data
@pytest.mark.skipif(bad_optional_imports,
                    reason="requires bokeh and astropy.stats.bls")
def test_full_widget():
    '''Test if we can run the widget with the keywords'''
    lc = KeplerLightCurve.read(KEPLER10)
    lc = lc.normalize().remove_nans().flatten()
    lc.interact_bls()
    lc.interact_bls(minimum_period=4)
    lc.interact_bls(maximum_period=5)
    lc.interact_bls(resolution=1000)


@pytest.mark.remote_data
@pytest.mark.skipif(bad_optional_imports,
                    reason="requires bokeh and astropy.stats.bls")
def test_tess_widget():
    '''Test if we can run the widget with the keywords'''
    lc = TessLightCurve.read(TESS_SIM)
    lc = lc.normalize().remove_nans().flatten()
    lc.interact_bls()
    lc.interact_bls(minimum_period=4)
    lc.interact_bls(maximum_period=5)
    lc.interact_bls(resolution=1000)



import pytest
import warnings

import numpy as np
from numpy.testing import assert_array_equal
import pandas as pd

from .. import DesignMatrix, DesignMatrixCollection
from ... import LightkurveWarning


def test_designmatrix_basics():
    """Can we create a design matrix from a dataframe?"""
    size, name = 10, 'testmatrix'
    df = pd.DataFrame({'vector1': np.ones(size),
                       'vector2': np.zeros(size),
                       'vector3': np.ones(size)})
    dm = DesignMatrix(df, name=name)
    assert dm.columns == ['vector1', 'vector2', 'vector3']
    assert dm.name == name
    assert dm.shape == (size, 3)
    assert (dm['vector1'] == df['vector1']).all()
    dm.plot()
    dm.plot_priors()
    assert dm.append_constant().shape == (size, 4)  # expect one column more
    assert dm.pca(nterms=2).shape == (size, 2)      # expect one column less
    assert dm.split([10]).shape == (size, 6)        # expect double columns
    dm.__repr__()

    dm = DesignMatrix(df, name=name)
    dm.append_constant(inplace=True)
    assert dm.shape == (size, 4)  # expect one column more

    dm = DesignMatrix(df, name=name)
    dm.split([10], inplace=True)
    assert dm.shape == (size, 6)        # expect double columns


def test_designmatrix_from_numpy():
    """Can we create a design matrix from an ndarray?"""
    size = 10
    dm = DesignMatrix(np.ones((size, 2)))
    assert dm.columns == [0, 1]
    assert dm.name == 'unnamed_matrix'
    assert (dm[0] == np.ones(size)).all()


def test_designmatrix_from_dict():
    """Can we create a design matrix from a dictionary?"""
    size = 10
    dm = DesignMatrix({'centroid_col': np.ones(size),
                       'centroid_row': np.ones(size)},
                      name='motion_systematics')
    assert dm.shape == (size, 2)
    assert (dm['centroid_col'] == np.ones(size)).all()


def test_split():
    """Can we split a design matrix correctly?"""
    dm = DesignMatrix({'a': np.linspace(0, 9, 10),
                       'b': np.linspace(100, 109, 10)})
    # Do we retrieve the correct shape?
    assert dm.shape == (10, 2)
    assert dm.split(2).shape == (10, 4)
    assert dm.split([2,8]).shape == (10, 6)
    # Are the new areas padded with zeros?
    assert (dm.split([2,8]).values[2:, 0:2] == 0).all()
    assert (dm.split([2,8]).values[:8, 4:] == 0).all()
    # Are all the column names unique?
    assert len(set(dm.split(2).columns)) == 4


def test_standardize():
    """Verifies DesignMatrix.standardize()"""
    # A column with zero standard deviation remains unchanged
    dm = DesignMatrix({'const': np.ones(10)})
    assert (dm.standardize()['const'] == dm['const']).all()
    # Normally-distributed columns will become Normal(0, 1)
    dm = DesignMatrix({'normal': np.random.normal(loc=5, scale=3, size=100)})
    assert np.round(np.median(dm.standardize()['normal']), 3) == 0
    assert np.round(np.std(dm.standardize()['normal']), 1) == 1
    dm.standardize(inplace=True)

def test_pca():
    """Verifies DesignMatrix.pca()"""
    size = 10
    dm = DesignMatrix({'a':np.random.normal(10, 20, size),
                       'b':np.random.normal(40, 10, size),
                       'c':np.random.normal(60, 5, size)})
    for nterms in [1, 2, 3]:
        assert dm.pca(nterms=nterms).shape == (size, nterms)


def test_collection_basics():
    """Can we create a design matrix collection?"""
    size = 5
    dm1 = DesignMatrix(np.ones((size, 1)), columns=['col1'], name='matrix1')
    dm2 = DesignMatrix(np.zeros((size, 2)), columns=['col2', 'col3'], name='matrix2')

    dmc = DesignMatrixCollection([dm1, dm2])
    assert_array_equal(dmc['matrix1'].values, dm1.values)
    assert_array_equal(dmc['matrix2'].values, dm2.values)
    assert_array_equal(dmc.values, np.hstack((dm1.values, dm2.values)))
    dmc.plot()
    dmc.__repr__()

    dmc = dm1.collect(dm2)
    assert_array_equal(dmc['matrix1'].values, dm1.values)
    assert_array_equal(dmc['matrix2'].values, dm2.values)
    assert_array_equal(dmc.values, np.hstack((dm1.values, dm2.values)))

    assert isinstance(dmc.to_designmatrix(), DesignMatrix)


def test_designmatrix_rank():
    """Does DesignMatrix issue a low-rank warning when justified?"""
    warnings.simplefilter("always")

    # Good rank
    dm = DesignMatrix({'a': [1, 2, 3]})
    assert dm.rank == 1
    dm.validate(rank=True)  # Should not raise a warning

    # Bad rank
    with pytest.warns(LightkurveWarning, match='rank'):
        dm = DesignMatrix({'a': [1, 2, 3], 'b': [1, 1, 1], 'c': [1, 1, 1],
                           'd': [1, 1, 1], 'e': [3, 4, 5]})
        assert dm.rank == 2
        dm.validate(rank=True) # Should raise a warning



"""Defines design matrix objects to aid linear regression problems.

Specifically, this module adds the `DesignMatrix`, `DesignMatrixCollection`,
`SparseDesignMatrix`, and `SparseDesignMatrixCollection` classes which
are design to work with the `RegressionCorrector` class.
"""
from copy import deepcopy
import warnings

import matplotlib.pyplot as plt
from numba import jit
import numpy as np
import pandas as pd
from scipy.sparse import lil_matrix, csr_matrix, hstack, vstack, issparse, find

from .. import MPLSTYLE
from ..utils import LightkurveWarning, plot_image


__all__ = ['DesignMatrix', 'SparseDesignMatrix',
           'DesignMatrixCollection', 'SparseDesignMatrixCollection']


class DesignMatrix():
    """A matrix of column vectors for use in linear regression.

    The purpose of this class is to provide a convenient method to interact
    with a set of one or more regressors which are known to correlate with
    trends or systematic noise signals which we want to remove from a light
    curve. Specifically, this class is designed to provide the design matrix
    for use by Lightkurve's `.RegressionCorrector` class.

    Parameters
    ----------
    df : dict, array, or `pandas.DataFrame` object
        Columns to include in the design matrix.  If this object is not a
        `~pandas.DataFrame` then it will be passed to the DataFrame constructor.
    columns : iterable of str (optional)
        Column names, if not already provided via ``df``.
    name : str
        Name of the matrix.
    prior_mu : array
        Prior means of the coefficients associated with each column in a linear
        regression problem.
    prior_sigma : array
        Prior standard deviations of the coefficients associated with each
        column in a linear regression problem.

    Examples
    --------
    >>> import lightkurve as lk
    >>> lk.DesignMatrix(np.arange(100), name='slope')
    slope DesignMatrix (100, 1)
    >>> lk.designmatrix.create_spline_matrix(np.arange(100), n_knots=5, name='spline')
    spline DesignMatrix (100, 5)
    """
    def __init__(self, df, columns=None, name='unnamed_matrix', prior_mu=None,
                 prior_sigma=None):
        if not isinstance(df, pd.DataFrame):
            df = pd.DataFrame(df)
        self.df = df
        if columns is not None:
            df.columns = columns
        self.columns = list(df.columns)
        self.name = name
        if prior_mu is None:
            prior_mu = np.zeros(len(df.T))
        if prior_sigma is None:
            prior_sigma = np.ones(len(df.T)) * np.inf
        self.prior_mu = np.atleast_1d(prior_mu)
        self.prior_sigma = np.atleast_1d(prior_sigma)

    @property
    def X(self):
        """Design matrix "X" to be used in RegressionCorrector objects"""
        return self.df.values

    def copy(self):
        """Returns a deepcopy of DesignMatrix"""
        return deepcopy(self)

    def plot(self, ax=None, **kwargs):
        """Visualize the design matrix values as an image.

        Uses Matplotlib's `~lightkurve.utils.plot_image` to visualize the
        matrix values.

        Parameters
        ----------
        ax : `~matplotlib.axes.Axes`
            A matplotlib axes object to plot into. If no axes is provided,
            a new one will be created.
        **kwargs : dict
            Extra parameters to be passed to `.plot_image`.

        Returns
        -------
        `~matplotlib.axes.Axes`
            The matplotlib axes object.
        """
        with plt.style.context(MPLSTYLE):
            ax = plot_image(self.values, ax=ax, xlabel='Component', ylabel='X',
                            clabel='Component Value', title=self.name,
                            interpolation='nearest', **kwargs)
            ax.set_aspect(self.shape[1]/(1.6*self.shape[0]))
            if self.shape[1] <= 40:
                ax.set_xticks(np.arange(self.shape[1]))
                ax.set_xticklabels([r'${}$'.format(i) for i in self.columns],
                                   rotation=90, fontsize=8)
        return ax

    def plot_priors(self, ax=None):
        """Visualize the coefficient priors.

        Parameters
        ----------
        ax : `~matplotlib.axes.Axes`
            A matplotlib axes object to plot into. If no axes is provided,
            a new one will be created.

        Returns
        -------
        `~matplotlib.axes.Axes`
            The matplotlib axes object.
        """
        def gauss(x, mu=0, sigma=1):
            return np.exp(-(x - mu)**2/(2*sigma**2))
        with plt.style.context(MPLSTYLE):
            if ax is None:
                _, ax = plt.subplots()
            for m, s in zip(self.prior_mu, self.prior_sigma):
                if ~np.isfinite(s):
                    ax.axhline(1, color='k')
                else:
                    x = np.linspace(m - 5*s, m + 5*s, 1000)
                    ax.plot(x, gauss(x, m, s), c='k')
            ax.set_xlabel('Value')
            ax.set_title('{} Priors'.format(self.name))
        return ax

    def _get_prior_sample(self):
        """Returns a random sample from the prior distribution."""
        return np.random.normal(self.prior_mu, self.prior_sigma)

    def split(self, row_indices, inplace=False):
        """Returns a new `.DesignMatrix` with regressors split into multiple
        columns.

        This method will return a new design matrix containing
        n_columns * len(row_indices) regressors.  This is useful in situations
        where the linear regression can be improved by fitting separate
        coefficients for different contiguous parts of the regressors.

        Parameters
        ----------
        row_indices : iterable of integers
            Every regressor (i.e. column) in the design matrix will be split
            up over multiple columns separated at the indices provided.

        Returns
        -------
        `.DesignMatrix`
            A new design matrix with shape (n_rows, len(row_indices)*n_columns).
        """
        if isinstance(row_indices, int):
            row_indices = [row_indices]
        if (len(row_indices) == 0) or (row_indices == [0]) or (row_indices is None):
            return self
        # Where do the submatrices begin and end?
        lower_idx = np.append(0, row_indices)
        upper_idx = np.append(row_indices, len(self.df))

        dfs = []
        for idx, a, b in zip(range(len(lower_idx)), lower_idx, upper_idx):
            new_columns = dict(
                ('{}'.format(val), '{}'.format(val) + ' {}'.format(idx + 1))
                for val in list(self.df.columns))
            dfs.append(self.df[a:b].rename(columns=new_columns))

        prior_mu = np.hstack([self.prior_mu for idx in range(len(dfs))])
        prior_sigma = np.hstack([self.prior_sigma for idx in range(len(dfs))])

        if inplace:
            dm = self
        else:
            dm = self.copy()
        dm.df = pd.concat(dfs, axis=1).fillna(0)
        dm.columns = dm.df.columns
        dm.prior_mu = prior_mu
        dm.prior_sigma = prior_sigma
        return dm

    def standardize(self, inplace=False):
        """Returns a new `.DesignMatrix` in which the columns have been
        median-subtracted and sigma-divided.

        For each column in the matrix, this method will subtract the median of
        the column and divide by the column's standard deviation, i.e. it
        will compute the column's so-called "standard scores" or "z-values".

        This operation is useful because it will make the matrix easier to
        visualize and makes fitted coefficients easier to interpret.

        Notes:
        * Standardizing a spline design matrix will break the splines.
        * Columns with constant values (i.e. zero standard deviation) will be
        left unchanged.

        Returns
        -------
        `.DesignMatrix`
            A new design matrix with median-subtracted & sigma-divided columns.
        """
        ar = np.asarray(np.copy(self.df))
        ar[ar == 0] = np.nan
        # If a column has zero standard deviation, it will not change!
        is_const = np.nanstd(ar, axis=0) == 0
        median = np.atleast_2d(np.nanmedian(ar, axis=0)[~is_const])
        std = np.atleast_2d(np.nanstd(ar, axis=0)[~is_const])
        ar[:, ~is_const] = (ar[:, ~is_const] - median) / std
        new_df = pd.DataFrame(ar, columns=self.columns).fillna(0)
        if inplace:
            dm = self
        else:
            dm = self.copy()
        dm.df = new_df
        return dm

    def pca(self, nterms=6):
        """Returns a new `.DesignMatrix` with a smaller number of regressors.

        This method will use Principal Components Analysis (PCA) to reduce
        the number of columns in the matrix.

        Parameters
        ----------
        nterms : int
            Number of columns in the new matrix.

        Returns
        -------
        `.DesignMatrix`
            A new design matrix with PCA applied.
        """
        # nterms cannot be langer than the number of columns in the matrix
        if nterms > self.shape[1]:
            nterms = self.shape[1]
        # We use `fbpca.pca` instead of `np.linalg.svd` because it is faster.
        # Note that fbpca is randomized, and has n_iter=2 as default,
        # we find this to be too few, and that n_iter=10 is still fast but
        # produces more stable results.
        from fbpca import pca  # local import because not used elsewhere
        new_values, _, _ = pca(self.values, nterms, n_iter=10)
        return DesignMatrix(new_values, name=self.name)

    def append_constant(self, prior_mu=0, prior_sigma=np.inf, inplace=False):
        """Returns a new `.DesignMatrix` with a column of ones appended.

        Returns
        -------
        `.DesignMatrix`
            New design matrix with a column of ones appended. This column is
            named "offset".
        """
        if inplace:
            dm = self
        else:
            dm = self.copy()
        extra_df = pd.DataFrame(np.atleast_2d(np.ones(self.shape[0])).T, columns=['offset'])
        dm.df = pd.concat([self.df, extra_df], axis=1)
        dm.columns = list(dm.df.columns)
        dm.prior_mu = np.append(self.prior_mu, prior_mu)
        dm.prior_sigma = np.append(self.prior_sigma, prior_sigma)
        return dm

    def _validate(self, rank=True):
        """Helper function for validating."""
        # Matrix rank shouldn't be significantly smaller than the # of columns
        if rank:
            if self.rank < (0.5*self.shape[1]):
                warnings.warn("The design matrix has low rank ({}) compared to the "
                              "number of columns ({}), which suggests that the "
                              "matrix contains duplicate or correlated columns. "
                              "This may prevent the regression from succeeding. "
                              "Consider reducing the dimensionality by calling the "
                              "`pca()` method.".format(self.rank, self.shape[1]),
                              LightkurveWarning)
        if self.prior_mu is not None:
            if len(self.prior_mu) != self.shape[1]:
                raise ValueError('`prior_mu` must have shape {}'
                                 ''.format(self.shape[1]))
        if self.prior_sigma is not None:
            if len(self.prior_sigma) != self.shape[1]:
                raise ValueError('`prior_sigma` must have shape {}'
                                 ''.format(self.shape[1]))
            if np.any(np.asarray(self.prior_sigma) <= 0):
                raise ValueError('`prior_sigma` values cannot be smaller than '
                                 'or equal to zero')

    def validate(self, rank=True):
        """Emits `LightkurveWarning` if matrix has low rank or priors have incorrect shape.

        Note that for `SparseDesignMatrix` objects, calculating the rank will
        force the design matrix to be evaluated and stored in memory, reducing
        the speed and memory savings of SparseDesignMatrix.

        For `SparseDesignMatrix`, rank checks will be turned off by default.
        """
        self._validate()

    @property
    def rank(self):
        """Matrix rank computed using `numpy.linalg.matrix_rank`."""
        return np.linalg.matrix_rank(self.values)

    @property
    def shape(self):
        """Tuple specifying the shape of the matrix as (n_rows, n_columns)."""
        return self.X.shape

    @property
    def values(self):
        """2D numpy array containing the matrix values."""
        return self.df.values

    def __getitem__(self, key):
        return self.df[key].values

    def __repr__(self):
        return '{} DesignMatrix {}'.format(self.name, self.shape)

    def to_sparse(self):
        """Convert this dense matrix object to a `SparseDesignMatrix`.

        The values of this design matrix will be converted to a
        `scipy.sparse.csr_matrix`, which stores the values in a
        lower memory matrix. This is not recommended for dense matrices.
        """
        return SparseDesignMatrix(csr_matrix(self.values),
                                  name=self.name,
                                  columns=self.columns,
                                  prior_mu=self.prior_mu,
                                  prior_sigma=self.prior_sigma)

    def collect(self, matrix):
        """ Join two designmatrices, return a design matrix collection """
        return DesignMatrixCollection([self, matrix])


class DesignMatrixCollection():
    """Object which stores multiple design matrices.

    DesignMatrixCollection objects are useful when users want to regress against
    multiple different systematics, but still keep the different systematics distinct.

    Examples
    --------
    >>> import lightkurve as lk
    >>> dm1 = lk.designmatrix.create_spline_matrix(np.arange(100), n_knots=5, name='spline')
    >>> dm2 = lk.DesignMatrix(np.arange(100), name='slope')
    >>> dmc = lk.DesignMatrixCollection([dm1, dm2])
    >>> dmc
    DesignMatrixCollection:
        spline DesignMatrix (100, 5)
        slope DesignMatrix (100, 1)
    >>> dmc.matrices
    [spline DesignMatrix (100, 5), slope DesignMatrix (100, 1)]
    """
    def __init__(self, matrices):
        if np.any([issparse(m.X) for m in matrices]):
            # This collection is designed for dense matrices, so we warn if a
            # SparseDesignMatrix is passed
            warnings.warn(('Some matrices are `SparseDesignMatrix` objects. '
                           'Sparse matrices will be converted to dense matrices.'),
                          LightkurveWarning)
            dense_matrices = []
            for m in matrices:
                if isinstance(m, SparseDesignMatrix):
                    dense_matrices.append(m.copy().to_dense())
                else:
                    dense_matrices.append(m)
            self.matrices = dense_matrices
        else:
            self.matrices = matrices
        self.X = np.hstack(tuple(m.X for m in self.matrices))
        self._child_class = DesignMatrix
        self.validate()

    @property
    def values(self):
        """2D numpy array containing the matrix values."""
        return np.hstack(tuple(m.values for m in self.matrices))

    @property
    def prior_mu(self):
        """Coefficient prior means."""
        return np.hstack([m.prior_mu for m in self])

    @property
    def prior_sigma(self):
        """Coefficient prior standard deviations."""
        return np.hstack([m.prior_sigma for m in self])

    def plot(self, ax=None, **kwargs):
        """Visualize the design matrix values as an image.

        Uses Matplotlib's `~lightkurve.utils.plot_image` to visualize the
        matrix values.

        Parameters
        ----------
        ax : `~matplotlib.axes.Axes`
            A matplotlib axes object to plot into. If no axes is provided,
            a new one will be created.
        **kwargs : dict
            Extra parameters to be passed to `.plot_image`.

        Returns
        -------
        `~matplotlib.axes.Axes`
            The matplotlib axes object.
        """
        temp_dm = DesignMatrix(pd.concat([d.df for d in self], axis=1))
        ax = temp_dm.plot(**kwargs)
        ax.set_title("Design Matrix Collection")
        return ax

    def plot_priors(self, ax=None):
        """Visualize the `prior_mu` and `prior_sigma` attributes.

        Parameters
        ----------
        ax : `~matplotlib.axes.Axes`
            A matplotlib axes object to plot into. If no axes is provided,
            a new one will be created.

        Returns
        -------
        `~matplotlib.axes.Axes`
            The matplotlib axes object.
        """
        [dm.plot_priors(ax=ax) for dm in self]
        return ax

    def _get_prior_sample(self):
        """Returns a random sample from the prior distribution."""
        return np.hstack([dm.sample_priors() for dm in self])

    def split(self, row_indices):
        """Returns a new `.DesignMatrixCollection` with regressors split into
        multiple columns.

        This method will return a new design matrix collection by calling
        `DesignMatrix.split` on each matrix in the collection.

        Parameters
        ----------
        row_indices : iterable of integers
            Every regressor (i.e. column) in the design matrix will be split
            up over multiple columns separated at the indices provided.

        Returns
        -------
        `.DesignMatrixCollection`
            A new design matrix collection.
        """
        return self.__class__([d.split(row_indices) for d in self])

    def standardize(self):
        """Returns a new `.DesignMatrixCollection` in which all the
        matrices have been standardized using the `DesignMatrix.standardize`
        method.

        Returns
        -------
        `.DesignMatrixCollection`
            The new design matrix collection.
        """
        return self.__class__([d.standardize() for d in self])

    @property
    def columns(self):
        """List of column names."""
        return np.hstack([d.columns for d in self])

    def __getitem__(self, key):
        try:
            return self.matrices[key]
        except Exception:
            arg = np.argwhere([m.name == key for m in self.matrices])
            return self.matrices[arg[0][0]]

    def validate(self):
        [d.validate() for d in self]

    def __repr__(self):
        return 'DesignMatrixCollection:\n' + \
                    ''.join(['\t{}\n'.format(i.__repr__()) for i in self])

    def to_designmatrix(self, name=None):
        """Flatten a `DesignMatrixCollection` into a `DesignMatrix`."""
        if name is None:
            name = self.matrices[0].name
        return self._child_class(self.X,
                                 columns=self.columns,
                                 prior_mu=self.prior_mu,
                                 prior_sigma=self.prior_sigma,
                                 name=name)


class SparseDesignMatrix(DesignMatrix):
    """A matrix of column vectors for use in linear regression.

    This class is similar to the `DesignMatrix` class, but uses the
    `scipy.sparse` library to improve speed in the case of sparse matrices.

    The purpose of this class is to provide a convenient method to interact
    with a set of one or more regressors which are known to correlate with
    trends or systematic noise signals which we want to remove from a light
    curve. Specifically, this class is designed to provide the design matrix
    for use by Lightkurve's `.RegressionCorrector` class.

    Parameters
    ----------
    X : `scipy.sparse` matrix
        The values to build the design matrix with
    columns : iterable of str (optional)
        Column names
    name : str
        Name of the matrix.
    prior_mu : array
        Prior means of the coefficients associated with each column in a linear
        regression problem.
    prior_sigma : array
        Prior standard deviations of the coefficients associated with each
        column in a linear regression problem.
    """
    def __init__(self, X, columns=None, name='unnamed_matrix', prior_mu=None,
                 prior_sigma=None):
        if not issparse(X):
            raise ValueError('Must pass a `scipy.sparse` matrix (e.g. `scipy.sparse.csr_matrix`)')
        if columns is None:
            columns = np.arange(X.shape[1])
        self.columns = columns
        self.name = name
        if prior_mu is None:
            prior_mu = np.zeros(X.shape[1])
        if prior_sigma is None:
            prior_sigma = np.ones(X.shape[1]) * np.inf
        self._X = X
        self.prior_mu = prior_mu
        self.prior_sigma = prior_sigma
        self._child_class = SparseDesignMatrix
        self.validate()

    @property
    def X(self):
        """Design matrix "X" to be used in RegressionCorrector objects"""
        return self._X

    @property
    def values(self):
        """2D numpy array containing the matrix values."""
        return self.X.toarray()

    def validate(self, rank=False):
        """Checks if the matrix has the right shapes. Set rank to True to test matrix rank."""
        # For sparse matrices, calculating the rank is expensive, and negates
        # the benefits of using sparse. Validate will ignore rank by default.
        self._validate(rank=rank)

    def split(self, row_indices, inplace=False):
        """Returns a new `.SparseDesignMatrix` with regressors split into multiple
        columns.

        This method will return a new design matrix containing
        n_columns * len(row_indices) regressors.  This is useful in situations
        where the linear regression can be improved by fitting separate
        coefficients for different contiguous parts of the regressors.

        Parameters
        ----------
        row_indices : iterable of integers
            Every regressor (i.e. column) in the design matrix will be split
            up over multiple columns separated at the indices provided.

        Returns
        -------
        `.SparseDesignMatrix`
            A new design matrix with shape (n_rows, len(row_indices)*n_columns).
        """

        if not hasattr(row_indices, '__iter__'):
            row_indices = [row_indices]

        # You can't split on the first or last index
        row_indices = list(np.asarray(row_indices)[~np.in1d(row_indices, [0, self.shape[0]])])
        if len(row_indices) == 0:
            return self
        if inplace:
            dm = self
        else:
            dm = self.copy()
        x = np.arange(dm.shape[0])
        dm.prior_mu = np.concatenate([list(self.prior_mu) * (len(row_indices) + 1)])
        dm.prior_sigma = np.concatenate([list(self.prior_sigma) * (len(row_indices) + 1)])
        dm._X = hstack([dm.X.multiply(lil_matrix(np.in1d(x, idx).astype(int)).T) for idx in np.array_split(x, row_indices)], format='lil')
        non_zero = dm.X.sum(axis=0) != 0
        non_zero = np.asarray(non_zero).ravel()
        dm._X = dm.X[:, non_zero]
        if dm.columns is not None:
            dm.columns = list(np.asarray([['{}_{}'.format(c, idx) for c in dm.columns] for idx in range(len(row_indices) + 1)]).ravel())
        dm.prior_mu = dm.prior_mu[non_zero]
        dm.prior_sigma = dm.prior_sigma[non_zero]
        return dm

    def standardize(self, inplace=False):
        """Returns a new `.SparseDesignMatrix` in which the columns have been
        mean-subtracted and sigma-divided.

        For each column in the matrix, this method will subtract the mean of
        the column and divide by the column's standard deviation, i.e. it
        will compute the column's so-called "standard scores" or "z-values".

        This operation is useful because it will make the matrix easier to
        visualize and makes fitted coefficients easier to interpret.

        Notes:
        * Standardizing a spline design matrix will break the splines.
        * Columns with constant values (i.e. zero standard deviation) will be
        left unchanged.

        Returns
        -------
        `.SparseDesignMatrix`
            A new design matrix with mean-subtracted & sigma-divided columns.
        """
        if inplace:
            dm = self
        else:
            dm = self.copy()

        idx, jdx, v = find(dm.X)
        weights = dm.X.copy()
        weights[dm.X != 0] = 1
        mean = np.bincount(jdx, weights=v)/np.bincount(jdx)
        std = np.asarray([((np.sum((v[jdx == i] - mean[i])**2) * (1/((jdx == i).sum() - 1))))**0.5 for i in np.unique(jdx)])
        mean[std == 0] = 0
        std[std == 0] = 1
        white = (dm.X - vstack([lil_matrix(mean)] * dm.shape[0])).multiply(vstack([lil_matrix(1/std)] * dm.shape[0]))
        dm._X = white.multiply(weights)
        return dm

    def pca(self, nterms=6, **kwargs):
        """Returns a new `.SparseDesignMatrix` with a smaller number of regressors.

        This method will use Principal Components Analysis (PCA) to reduce
        the number of columns in the matrix.

        Parameters
        ----------
        nterms : int
            Number of columns in the new matrix.

        Returns
        -------
        `.SparseDesignMatrix`
            A new design matrix with PCA applied.
        """
        return super().pca(nterms, **kwargs).to_sparse()

    def append_constant(self, prior_mu=0, prior_sigma=np.inf, inplace=False):
        """Returns a new `.SparseDesignMatrix` with a column of ones appended.

        Returns
        -------
        `.SparseDesignMatrix`
            New design matrix with a column of ones appended. This column is
            named "offset".
        """
        if inplace:
            dm = self
        else:
            dm = self.copy()
        dm._X = hstack([dm.X, lil_matrix(np.ones(dm.shape[0])).T], format='lil')
        dm.prior_mu = np.append(dm.prior_mu, prior_mu)
        dm.prior_sigma = np.append(dm.prior_sigma, prior_sigma)
        return dm

    def __getitem__(self, key):
        loc = np.where(np.asarray(self.columns) == key)[0]
        if len(loc) == 0:
            raise ValueError('No such column as `{}`.'.format(key))
        return self.X[:, loc].toarray()

    def __repr__(self):
        return '{} SparseDesignMatrix {}'.format(self.name, self.shape)

    def collect(self, matrix):
        """ Join two designmatrices, return a design matrix collection """
        return SparseDesignMatrixCollection([self, matrix])

    def to_dense(self):
        """ Convert a SparseDesignMatrix object to a dense DesignMatrix

        The values of this design matrix will be converted to a
        `numpy.ndarray`. This is not recommended for sparse matrices containing
        mostly zeros.
        """
        return DesignMatrix(self.values, name=self.name,
                            columns=self.columns, prior_mu=self.prior_mu,
                            prior_sigma=self.prior_sigma)


class SparseDesignMatrixCollection(DesignMatrixCollection):
    """A set of design matrices."""
    def __init__(self, matrices):
        if not np.all([issparse(m.X) for m in matrices]):
            # This collection is designed for sparse matrices, so we raise a warning if a dense DesignMatrix is passed
            warnings.warn(('Not all matrices are `SparseDesignMatrix` objects. '
                            'Dense matrices will be converted to sparse matrices.'), LightkurveWarning)
            sparse_matrices = []
            for m in matrices:
                if isinstance(m, DesignMatrix):
                    sparse_matrices.append(m.copy().to_sparse())
                else:
                    sparse_matrices.append(m)
            self.matrices = sparse_matrices
        else:
            self.matrices = matrices
        self.X = hstack([m.X for m in self.matrices], format='csr')
        self._child_class = SparseDesignMatrix
        self.validate()

    def plot(self, ax=None, **kwargs):
        """Visualize the design matrix values as an image.

        Uses Matplotlib's `~lightkurve.utils.plot_image` to visualize the
        matrix values.

        Parameters
        ----------
        ax : `~matplotlib.axes.Axes`
            A matplotlib axes object to plot into. If no axes is provided,
            a new one will be created.
        **kwargs : dict
            Extra parameters to be passed to `.plot_image`.

        Returns
        -------
        `~matplotlib.axes.Axes`
            The matplotlib axes object.
        """
        temp_dm = SparseDesignMatrix(hstack([d.X for d in self]))
        ax = temp_dm.plot(**kwargs)
        ax.set_title("Design Matrix Collection")
        return ax

    def __repr__(self):
        return 'SparseDesignMatrixCollection:\n' + \
                    ''.join(['\t{}\n'.format(i.__repr__()) for i in self])



####################################################
# Functions to create commonly-used design matrices.
####################################################

@jit(nopython=True)
def _spline_basis_vector(x, degree, i, knots):
    """Recursive function to create a single spline basis vector for an input x,
    for the ith knot.

    See https://en.wikipedia.org/wiki/B-spline for a definition of B-spline
    basis vectors

    Parameters
    ----------
    x : np.ndarray
        Input x
    degree : int
        Degree of spline to calculate basis for
    i : int
        The index of the knot to calculate the basis for
    knots : np.ndarray
        Array of all knots

    Returns
    -------
    B : np.ndarray
        A vector of same length as x containing the spline basis for the ith knot
    """
    if degree == 0:
        B = np.zeros(len(x))
        B[(x >= knots[i]) & (x <= knots[i+1])] = 1
    else:
        da = (knots[degree + i] - knots[i])
        db = (knots[i + degree + 1] - knots[i + 1])
        if ((knots[degree + i] - knots[i]) != 0):
            alpha1 = ((x - knots[i])/da)
        else:
            alpha1 = np.zeros(len(x))
        if ((knots[i+degree+1] - knots[i+1]) != 0):
            alpha2 = ((knots[i + degree + 1] - x)/db)
        else:
            alpha2 = np.zeros(len(x))
        B = (_spline_basis_vector(x, (degree-1), i, knots)) * (alpha1) + (_spline_basis_vector(x, (degree-1), (i+1), knots)) * (alpha2)
    return B


def create_sparse_spline_matrix(x, n_knots=20, knots=None, degree=3, name='spline'):
    """Creates a piecewise polynomial function, creating a continuous, smooth function in x

    See https://en.wikipedia.org/wiki/B-spline for the definitions of Basis Splines

    B-spline vectors of degree higher than 0 are created using recursion, using the
    `_spline_basis_vector` function to evaluate the basis vectors for x, for each knot.

    Parameters
    ----------
    x : np.ndarray
        vector to spline
    n_knots: int
        Number of knots (default: 20).
    knots : np.ndarray [optional]
        Optional array containing knots
    degree: int
        Polynomial degree.
    name: string
        Name to pass to `.SparseDesignMatrix` (default: 'spline').
    include_intercept: bool
        Whether to include row of ones to find intercept. Default False.

    Returns
    -------
    dm: `.SparseDesignMatrix`
        Design matrix object with shape (len(x), n_knots*degree).
    """
    # To use jit we have to use float64
    x = np.asarray(x, np.float64)

    if not isinstance(n_knots, int):
        raise ValueError('`n_knots` must be an integer.')
    if n_knots - degree <= 0:
        raise ValueError('n_knots must be greater than degree.')
    if (knots is None)  and (n_knots is not None):
        knots = np.asarray([s[-1] for s in np.array_split(np.argsort(x), n_knots - degree)[:-1]])
        knots = [np.mean([x[k], x[k + 1]]) for k in knots]
    elif (knots is None)  and (n_knots is None):
        raise ValueError('Pass either `n_knots` or `knots`.')
    knots = np.append(np.append(x.min(), knots), x.max())
    knots = np.unique(knots)
    knots_wbounds = np.append(np.append([x.min()] * (degree - 1), knots), [x.max()] * (degree))

    matrices = [csr_matrix(_spline_basis_vector(x, degree, idx, knots_wbounds)) for idx in np.arange(-1, len(knots_wbounds) - degree - 1)]
    spline_dm = vstack([m for m in matrices if (m.sum() != 0) ], format='csr').T
    return SparseDesignMatrix(spline_dm, name=name)


def create_spline_matrix(x, n_knots=20, knots=None, degree=3, name='spline',
                         include_intercept=True):
    """Returns a `.DesignMatrix` which models splines using `patsy.dmatrix`.
    Parameters
    ----------
    x : np.ndarray
        vector to spline
    n_knots: int
        Number of knots (default: 20).
    degree: int
        Polynomial degree.
    name: string
        Name to pass to `.DesignMatrix` (default: 'spline').
    include_intercept: bool
        Whether to include row of ones to find intercept. Default False.
    Returns
    -------
    dm: `.DesignMatrix`
        Design matrix object with shape (len(x), n_knots*degree).
    """
    from patsy import dmatrix  # local import because it's rarely-used
    if knots is not None:
        dm_formula = "bs(x, knots={}, degree={}, include_intercept={}) - 1" \
                     "".format(knots, degree, include_intercept)
        spline_dm = np.asarray(dmatrix(dm_formula, {"x": x}))
        df = pd.DataFrame(spline_dm, columns=['knot{}'.format(idx + 1)
                                              for idx in range(spline_dm.shape[1])])
    else:
        dm_formula = "bs(x, df={}, degree={}, include_intercept={}) - 1" \
                 "".format(n_knots, degree, include_intercept)
        spline_dm = np.asarray(dmatrix(dm_formula, {"x": x}))
        df = pd.DataFrame(spline_dm, columns=['knot{}'.format(idx + 1)
                                              for idx in range(n_knots)])
    return DesignMatrix(df, name=name)
